# This is the main initialization file for all Soar interpreters.
# This is the first Tcl code that is executed in an interpreter
# after it is created.  This code is loaded before any user-specific
# initialization file, such as the <interpreter name>.soar files.

global interp_type
global interp_name

#
# When an alias is defined, a new procedure is created at the
# top level whose name matches the alias.  The string defining
# the alias is used as the prefix of the actual command issued.
# If the alias command is invoked with additional arguments
# these are post-pended onto the command prefix.  Note the
# argument "args" is treated like Common Lisp's &rest argument.
#

global defined_aliases
set defined_aliases {}

global print_alias_switch
set print_alias_switch "off"

proc alias {name args} {
  global defined_aliases print_alias_switch

  if {[string compare $print_alias_switch "on"] == 0} {
      echo "Defining alias \"$name\" as \"$args\""
  }

  if {[lsearch -exact [info commands] $name] >= 0} {
      if {[lsearch -exact $defined_aliases $name] < 0} {
	  echo "Warning: the alias \"$name\" has replaced a previously defined command."
#      }
  }

  set defined_aliases [linsert $defined_aliases 0 $name]

  uplevel #0 "proc $name {args} {
                  if {\$args == \"\"} {
                    $args 
                  } else {
                   eval $args \$args
		  }
	      }"
}

proc print-alias {{setting ""}} {
  global print_alias_switch defined_aliases

  if {[string compare $setting ""] == 0} {
      return $print_alias_switch
  }

  switch -exact -- $setting {
      -on     {set print_alias_switch "on"}
      -off    {set print_alias_switch "off"}
      -list   {lsort [set defined_aliases]}
      default {error "Unrecognized argument: $setting.  Usage: print-alias \[-on | -off]."}
  }
}

proc unalias {name} {
    global defined_aliases
    set position [lsearch -exact $defined_aliases $name]
    if {$position >= 0} {
	set defined_aliases [lreplace $defined_aliases $position $position]
	rename $name {}
    } else {
	echo Error: There is no alias named \"$name\".
    }
}

#---------------------------------------------------------------

# These alias definitions replace C code in the pre-Tcl 
# interpreter.

if [string match $interp_type agent] {
    alias ?     help
    alias wmem  print -depth 0 -internal

# More complex aliases can be defined using proc.  Note that
# the argument syntax {<arg> <initial value if unspecified>}
# is treated like Common Lisp's &optional argument.

    proc d {{num -1}} {
	go d $num
    }

    proc run {{num -1}} {
	go e $num
    }
}

#---------------------------------------------------------------
#
# Help procedures

global documented_aliases
set documented_aliases "? wmem create-agent destroy-agent \
excise-all excise-chunks excise-task list-agents list-chunks \
list-help-topics list-justifications load memory-stats \
object-trace-format p pgs print-stats rete-stats r select-agent \
spr stack-trace-format"

global pager_cache
set pager_cache ""

global help_no_args_string
set help_no_args_string "
Commonly used Soar commands:
  cd		Change to another directory
  excise	Remove productions from Soar's memory
  go		Run Soar
  help		Get help on a specific command
  init-soar	Reinitialize Soar
  learn		Set learning on and off
  log		Save a Soar session to a file
  matches	List information about partial matches for a production
  ms		List productions currently in the match set
  preferences	List all the preferences for a given object and attribute
  print		Display a Soar object (production or  working memory elements)
  pwd		Display the current working directory
  quit		Exit Soar
  soarnews	Display information such as where to report bugs
  source	Load a file into Soar
  sp		Define a Soar production
  user-select	Set the method for arbitrating indifferent preference ties. 
  version	Display the version number of Soar
  watch		Set the amount of information displayed as Soar runs
  wmem		Display working memory elements

For a list of ALL available help topics, type \"help -all\"

For help on a specific command, type \"help\" followed by the command name.
"

proc help args {
    global soar_library pager_cache env help_no_args_string 
    global documented_aliases tcl_man_dir tk_man_dir

    if {([llength $args] == 0) || ([lindex $args 0] == "-beginner")} {
	return $help_no_args_string
    }

    if {[lindex $args 0] == "-all"} {
	set d [pwd]
	cd $soar_library/../doc/man
	set topics [glob -nocomplain -- *.n]
	cd $tcl_man_dir
        set topics [concat $topics [glob -nocomplain -- *.n]]
	if [info exists tk_man_dir] {
	    cd $tk_man_dir
	    set topics [concat $topics [glob -nocomplain -- *.n]]
	}
	set names ""
	foreach f $topics {
	    set names [concat $names [file rootname $f]]
	}
	cd $d
	return [lsort $names]
    }

    set unix_man_page 0

    if {[lindex $args 0] == "-usage"} {
	if {[llength $args] > 1} {
	    set topic [lindex $args 1]
	    if {[file exists $soar_library/../doc/man/$topic.n]} {
		set dir $soar_library/../doc/man
	    } elseif {[file exists $tcl_man_dir/$topic.n]} {
		set dir $tcl_man_dir
	    } elseif {[file exists $tk_man_dir/$topic.n]} {
		set dir $tk_man_dir
	    } else {

# See if its an alias

                if {[lsearch $documented_aliases $topic] >= 0} {
		    return "\"$topic\" is an alias, not a builtin command."
		} else {

# One last try.  See if its in a man page.

                    echo "\"$topic\" is not a Soar, Tk, or Tcl command."
                    echo "Checking to see if it is a Unix command..."
                    set text [exec man $topic]
                    if [regexp "SYNOPSIS" $text] {
		        set unix_man_page 1
		    } else {
			return "No help is available for the topic \"$topic\".  \"help -all\" will list\nall available help topics and \"man\" gives help on Unix commands."
		    }
		}
	    }

	    if {!$unix_man_page} {
		set d [pwd]
		cd $dir
		set f [open "| nroff -man $topic.n" r]
	    } else {
		set f [open "| man $topic" r]
	    }
	    while {![eof $f]} {
		gets $f line
		if [regexp {^[ ]*(S.)*S(Y.)*Y(N.)*N(O.)*O(P.)*P(S.)*S(I.)*I(S.)*S} \
			"$line"] {
		    gets $f line
		    while {   ![regexp {___} "$line"] 
                           && ![string match "" "$line"]
		           && ![eof $f]} {
			echo $line
			gets $f line
		    }
		    break
		}
	    }
	    catch "close $f"
	    return
	} else {
	    return "help -usage: No topic given."
	}
    }

    set topic [lindex $args 0]

    if {[lsearch $documented_aliases $topic] >= 0} {
        set dir $soar_library/../doc/man
        set topic aliases
    } elseif {[file exists $soar_library/../doc/man/$topic.n]} {
	set dir $soar_library/../doc/man
    } elseif {[file exists $tcl_man_dir/$topic.n]} {
	set dir $tcl_man_dir
    } elseif {[file exists $tk_man_dir/$topic.n]} {
	set dir $tk_man_dir
    } else {

# One last try.  See if its in a man page.

        echo "\"$topic\" is not a Soar, Tk, or Tcl command."
        echo "Checking to see if it is a Unix command..."
        set text [exec man $topic]
        if [regexp "SYNOPSIS" $text] {
	    set unix_man_page 1
	} else {
	    return "No help is available for the topic \"$topic\".  \"help -all\" will list\nall available help topics and \"man\" gives help on Unix commands."
	}
    }

    if {!$unix_man_page} {
	set d [pwd]
	cd $dir
    }

#   Find pager to use    

    set pager $pager_cache
    if {$pager == ""} {
	catch {set pager $env(PAGER)}
	if {$pager == ""} {
	    catch {exec more} msg
	    if [regexp {couldn't find} $msg] {
		catch {exec page} msg
		if [regexp {couldn't find} $msg] {
		    catch {exec less} msg
		    if [regexp {couldn't find} $msg] {
		    } else {
			set pager less
		    }			
		} else {
		    set pager page
		}
	    } else {
		set pager more
	    }
	}
    }

    if {$pager != ""} {
	set pager_cache $pager
	set tempfile /tmp/soartk.out[pid]
	if {$unix_man_page} {
	    exec man $topic >$tempfile
	} else {
	    exec nroff -man $topic.n >$tempfile
	}
	uplevel #0 $pager $tempfile
    } else {
	echo "Cannot find pager to show help.  Please set the env. var. PAGER."
    }

    if {!$unix_man_page} {
	cd $d
    }

    return
}

#---------------------------------------------------------------

# Setup exit and quit to execute both Soar and Tcl exit cleanup
# procedures.

if [string match $interp_type agent] {
    rename exit tcl-exit
    alias exit quit
} else {
    proc quit {} {
	exit
    }
}

#---------------------------------------------------------------
#
# Include pushd, popd, and dirs commands from the TclX extension.

source $soar_library/pushd.tcl

#---------------------------------------------------------------

# Redefine the default prompt:

global tcl_prompt1

if [string match $interp_type agent] {
    set tcl_prompt1 "puts -nonewline stdout \"$interp_type $interp_name> \"; log -add \"$interp_type $interp_name> \""
} else {
    set tcl_prompt1 "puts -nonewline stdout \"$interp_type $interp_name> \""
}

#---------------------------------------------------------------

# Interpreter-specific initializations.  This code checks for an 
# initialization file using the path specified in the command line
# arguments.  If the path is not defined, then . and $HOME are
# considered.
#
# If found, the init file is sourced into the interpreter to 
# initialize its behavior.

proc init-interpreter {} {
    global interp_name path env

    if {$path == ""} {
	set path ".:$env(HOME)"
    }

# Attempt to find the init file by scanning the given paths
    
    foreach dir [split $path :] {
	if [file exists $dir/$interp_name.soar] {
	    pushd $dir
	    source $interp_name.soar
	    popd
	    return
	}
    }
}

#---------------------------------------------------------------

# Helpful procedures.

# Dump the printed output and results of a command to a file

if [string match $interp_type agent] {

    proc command-to-file {command arg1 {arg2 ""}} {
	if {[string match $arg2 "" ]} {
	    set file $arg1
	    set mode w
	} elseif {[string match $arg1 "-new"]} {
	    set file $arg2
	    set mode w
	} elseif {[string match $arg2 "-new"]} {
	    set file $arg1
	    set mode w
	} elseif {[string match $arg1 "-existing"]} {
	    set file $arg2
	    set mode a
	} elseif {[string match $arg2 "-existing"]} {
	    set file $arg1
	    set mode a
	} else {
	    error "command-to-file: Unrecognized arguments: $arg1 $arg2"
	}
	set f [open $file $mode]
	output-strings-destination -push -file-id $f
	uplevel #0 [echo [eval $command]]
	output-strings-destination -pop
	close $f
    }
}

# Evaluate a command in a list of interpreters

proc send-to-interpreters {command interp_list} {
    foreach int $interp_list {
	send $int "$command"
    }
}
