
/*
 *----------------------------------------------------------------------
 *
 * my_input_function --
 *
 *      This function adds input to Soar's working memory.  This
 *      function serves as an example of how to add such input to Soar.
 *      This example simply increments a counter and places the
 *      counter value on Soar's input-link.  The structure that
 *      is created is defined above.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Adds elements to Soar's working memory.  May also define
 *      new symbols.
 *
 *----------------------------------------------------------------------
 */


#include "robocup.h"
#include "soar2c.h"
#include <soar.h>

#include "intercept.h"
#include "cmd_timing.h"

extern void robocup_init();
extern int receive_and_parse();
extern get_word(char *, int, char *);

/* Unfortunately, due to all kinds of considerations, the only practical way
   of doing this in an orderly fashion is to have many many global variables.
   These are listed below.  At least we made them static to prevent accidental
   use from outside the file  */

/* top-state input link */
static Symbol *top_input_link_id;  /* needed as root to all wme's on the input */

/* attribute names, values, and wme vars */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* common attribute names */
static Symbol *no_symbol;           /* "*no*"   */
static Symbol *yes_symbol;          /* "*yes*"  */
static Symbol *distance_symbol;     /* "distance"  */
static Symbol *angle_symbol;        /* "angle"  */
static Symbol *visible_symbol;      /* "visible"  */
static Symbol *power_symbol;        /* "power"  */

/* recommendation for kick angle */
static Symbol *kick_angle_symbol;
static Symbol *kick_angle_value;
static wme    *kick_angle_wme;

/* recommendation for c4.5 kicking angle */
static Symbol *smart_kick_symbol;
static Symbol *smart_kick_value;
static wme    *smart_kick_wme;

/* recommendation for clearing the ball angle */
static Symbol *clear_ball_symbol;
static Symbol *clear_ball_value;
static wme    *clear_ball_wme;

/* interception */
static Symbol *intercept_angle_value_sym;
static Symbol *intercept_power_value_sym;
static Symbol *intercept_obj_id;
static wme    *intercept_angle_wme, *intercept_power_wme;

/* server time */
static Symbol *value_sym;                  /* value */ 
static Symbol *time_obj_id, *new_int_sym;
static wme    *value_wme;

/* ball */
static Symbol *ball_obj_id, 
  *ball_distance_value_sym, *ball_angle_value_sym,
  *ball_other_close_sym, *ball_us_close_sym;

static wme    *ball_visible_wme, /* visibility */
  *new_ball_dist_wme, 
  *new_ball_angle_wme,
  *new_ball_other_close_wme,
  *new_ball_us_close_wme;

/* goals */
typedef struct {
  Symbol *obj_id;

  Symbol *distance_value;
  Symbol *angle_value;

  wme *visible_wme;
  wme *distance_wme;
  wme *angle_wme;
} goal_soar_struct;

static goal_soar_struct goals[TEAMS];  /* robocup.h */

/* flags */
typedef struct {
  Symbol *obj_id;
  Symbol *distance_value;
  Symbol *angle_value;

  wme    *visible_wme;
  wme    *distance_wme;
  wme    *angle_wme;
} flag_soar_struct;

static flag_soar_struct flags[NUM_FLAGS];  /* robocup.h: # of enumerations in FlagPos in libsclient.h */ 

/* players and team */
typedef struct {
  Symbol *obj_id;
  Symbol *distance_value;
  Symbol *angle_value;
  Symbol *direction_value;
  Symbol *absolute_position_obj_id;
  Symbol *absolute_position_x_value;
  Symbol *absolute_position_y_value;
  Symbol *absolute_position_theta_value;

  wme    *visible_wme;
  wme    *distance_wme;
  wme    *angle_wme;
  wme    *direction_wme;
  wme    *absolute_position_x_wme;
  wme    *absolute_position_y_wme;
  wme    *absolute_position_theta_wme;
} player_soar_struct;

static player_soar_struct players[TEAMS][MAX_PLAYERS];  /* robocup.h */
static Symbol *player_direction_sym_const;
static Symbol *player_x_symbol;
static Symbol *player_y_symbol;
static Symbol *player_theta_symbol;

/* us_close, other_close, # of people closer to ball */
typedef struct {
  Symbol *close_symbol;
  Symbol *close_value;
  wme    *close_wme;
} close_count_struct;

static close_count_struct close_count[TEAMS];

/* direction and power to go back to home position */
static Symbol *home_dir_obj_id;
static Symbol *home_dir_value, *home_power_value;
static wme *home_dir_wme, *home_power_wme;

/*******************************************
  initialize_top_state
********************************************/

static void initialize_top_state (agent *a)
{
  robocup_init();        /* initialize socket, state, and parser */
  intercept_init();      /* initialize parameters */

  /*
    other initialization functions should be here
   */

  /* set common attritute names */
  no_symbol = get_io_sym_constant ("*no*"); 
  yes_symbol= get_io_sym_constant ("*yes*");
  distance_symbol= get_io_sym_constant ("distance");
  angle_symbol= get_io_sym_constant ("angle");
  visible_symbol= get_io_sym_constant ("visible");
  power_symbol  = get_io_sym_constant ("power");

  /* create and initialize input link - must be used before any and all CREATE_WMEs */
  {
    Symbol *input_link_sym;

    /* The following creates the wme (state.io ^input-link L<something>) in WM */
    /* 
       input_link_sym = get_io_sym_constant ("input-link");
       top_input_link_id = get_new_io_identifier ('L');
       top_input_link_wme = add_input_wme (a->io_header, input_link_sym, top_input_link_id);
    */

    /* throwing away pointer to the new wme */
    (void) CREATE_WME(a->io_header,  
		      input_link_sym, get_io_sym_constant ("input-link"),
		      top_input_link_id, get_new_io_identifier ('L'));
		     
    release_io_symbol (input_link_sym);
  }

  /* timing init needs the input link */
  cmd_timing_init(top_input_link_id);

  
  /* kick angle recommendation */
  {
    kick_angle_wme = CREATE_WME(top_input_link_id,
				kick_angle_symbol, get_io_sym_constant("kick_angle"),
				kick_angle_value, get_io_float_constant(0.0));
  }

  /* smart kicking angle recommendation: c4.5 kicking */
  {
    extern init(char *);
    
    init("./DF");
    smart_kick_wme = CREATE_WME(top_input_link_id,
				smart_kick_symbol, get_io_sym_constant("smart_kick_angle"),
				smart_kick_value, get_io_float_constant(0.0));
  }
 
  /* clearing the ball angle recommendation */
  {
    clear_ball_wme = CREATE_WME(top_input_link_id,
				clear_ball_symbol, get_io_sym_constant("clear_ball_angle"),
				clear_ball_value, get_io_float_constant(0.0));
  }
 
  /* interception initialization */
  {
    Symbol *intercept_sym;
    
    (void) CREATE_WME(top_input_link_id,
		      intercept_sym, get_io_sym_constant ("intercept"),
		      intercept_obj_id, get_new_io_identifier ('C'));
 
    release_io_symbol (intercept_sym);
 
    intercept_angle_wme = CREATE_WME_VALUE(intercept_obj_id, angle_symbol,
					   intercept_angle_value_sym, get_io_float_constant(0.0));

    intercept_power_wme = CREATE_WME_VALUE(intercept_obj_id, power_symbol,
				     intercept_power_value_sym, get_io_int_constant(0));
  }

  /* server time */
  {
    Symbol *time_sym;

    (void) CREATE_WME(top_input_link_id,
		      time_sym, get_io_sym_constant ("time"),
		      time_obj_id, get_new_io_identifier ('T'));
    
    release_io_symbol (time_sym);

    value_wme = CREATE_WME(time_obj_id,
			   value_sym, get_io_sym_constant ("value"),
			   new_int_sym, get_io_int_constant ((int)0));
    
  }

  /* ball */
  {
    Symbol *ball_sym;

    /* link to ball information */
    (void) CREATE_WME(top_input_link_id,
		      ball_sym, get_io_sym_constant ("ball"),
		      ball_obj_id, get_new_io_identifier ('B'));

    release_io_symbol (ball_sym);

    /* ball visibility */
    ball_visible_wme = CREATE_WME_CONST(ball_obj_id, visible_symbol, no_symbol);

    /* ball distance */
    new_ball_dist_wme = CREATE_WME_VALUE(ball_obj_id, distance_symbol,
					 ball_distance_value_sym, get_io_float_constant(-1.0));

    /* ball angle */
    new_ball_angle_wme = CREATE_WME_VALUE(ball_obj_id, angle_symbol,
				    ball_angle_value_sym, get_io_float_constant(0.0));

    /* ball other-close */
    new_ball_other_close_wme = CREATE_WME_ATTRIB_CONST(ball_obj_id,
					  ball_other_close_sym, get_io_sym_constant("other_close"),
					  no_symbol);

    /* ball us-close */
    new_ball_us_close_wme = CREATE_WME_ATTRIB_CONST(ball_obj_id,
				       ball_us_close_sym, get_io_sym_constant("us_close"),
				       no_symbol);

    /* ball friendlies_closer */
    close_count[OWN].close_wme = CREATE_WME(ball_obj_id,
					    close_count[OWN].close_symbol,get_io_sym_constant("close-friendlies"),
					    close_count[OWN].close_value, get_io_int_constant(0));

    /* ball opponents closer */
    close_count[OPPONENT].close_wme = CREATE_WME(ball_obj_id,
						 close_count[OPPONENT].close_symbol,get_io_sym_constant("close-opponents"),
						 close_count[OPPONENT].close_value, get_io_int_constant(0));
  }

  /* goals */
  {
    Symbol *goal_sym;
    int i;
    char *goal_names[TEAMS] = { "my_goal", "other_goal" }; /* XXX assumes OWN = 0, OPPONENT =1 */

    for (i=0; i<TEAMS; i++)
      {
	(void) CREATE_WME(top_input_link_id, 
			  goal_sym, get_io_sym_constant (goal_names[i]),
			  goals[i].obj_id, get_new_io_identifier ('G'));

	release_io_symbol(goal_sym);

	goals[i].visible_wme = CREATE_WME_CONST(goals[i].obj_id, visible_symbol, no_symbol);

	goals[i].distance_wme = CREATE_WME_VALUE(goals[i].obj_id, distance_symbol,
						 goals[i].distance_value, get_io_float_constant(-1.0));
	
	goals[i].angle_wme = CREATE_WME_VALUE(goals[i].obj_id, angle_symbol,
					      goals[i].angle_value, get_io_float_constant(0.0));
      }
  }

  /* flag information */
  {
    Symbol *flag_sym, *flag_type_sym, *flag_type_value_sym;

    /* XXX keep this variable synchronized with the enumerations of FlagPos in libsclient.h!  */
    char *flag_types[NUM_FLAGS]  = { 
      "unknown",
      "left-top",
      "left-bottom",
      "right-top",
      "right-bottom",
      "top", /* center top */
      "bottom", /* center bottom */
      "penalty-left-top",
      "penalty-left-center",
      "penalty-left-bottom",
      "penalty-right-top",
      "penalty-right-center",
      "penalty-right-bottom"
    };
    int i;

    flag_sym = get_io_sym_constant ("flag");
    flag_type_sym = get_io_sym_constant ("type");

    for (i=0; i<NUM_FLAGS; i++)
      { /* for each flag */
   
	/* link flag to input */
	(void) CREATE_WME_VALUE(top_input_link_id, flag_sym, 
			  flags[i].obj_id, get_new_io_identifier ('F'));
    
	/* set flag type */
	(void) CREATE_WME_VALUE(flags[i].obj_id, flag_type_sym,
			  flag_type_value_sym, get_io_sym_constant(flag_types[i]));

	release_io_symbol(flag_type_value_sym);
 
	/* flag visibility */
	flags[i].visible_wme = CREATE_WME_CONST(flags[i].obj_id, 
						visible_symbol,
						no_symbol);

	/* flag angles */
	flags[i].angle_wme   = CREATE_WME_VALUE(flags[i].obj_id, angle_symbol,
						flags[i].angle_value, get_io_float_constant(0.0));

	/* flag distances */
	flags[i].distance_wme   = CREATE_WME_VALUE(flags[i].obj_id, distance_symbol,
					     flags[i].distance_value, get_io_float_constant(-1.0));
      }

    /* Release symbols we won't need any more */

    release_io_symbol (flag_sym);
    release_io_symbol (flag_type_sym);
  }

  /* players and teams */
  {
    Symbol *player_sym;
    Symbol *player_team_sym;
    Symbol *player_number_sym;
    Symbol *team_number_sym;
    Symbol *number_value;
    Symbol *player_position_sym;

    int team, player;  /* indexes to team and players in array */
    
    player_sym           = get_io_sym_constant ("player");
    player_team_sym      = get_io_sym_constant ("team");
    player_number_sym    = get_io_sym_constant ("number");
    player_position_sym  = get_io_sym_constant ("absolute-position");

    /* initialize constants that will be used later */
    player_x_symbol      = get_io_sym_constant ("x");
    player_y_symbol      = get_io_sym_constant ("y");
    player_theta_symbol  = get_io_sym_constant ("theta");

    /* direction is different, since it is used later on in the update loop */
    player_direction_sym_const = get_io_sym_constant ("direction");     

    /* init only our own players.  We want be looking for other opponents */
    for (team=0; team<TEAMS; team++)
      {
	team_number_sym = get_io_int_constant(team);

	for (player=0; player<MAX_PLAYERS; player++)
	  {
	    /* create the player wme off the input-link */
	    (void) CREATE_WME_VALUE(top_input_link_id, player_sym, 
			      (players[team][player]).obj_id, get_new_io_identifier ('P'));

	    /* player number */
	    (void) CREATE_WME_VALUE((players[team][player]).obj_id, player_number_sym, 
			      number_value, get_io_int_constant(player+1));
	    release_io_symbol(number_value);

	    /* player team */
	    (void) CREATE_WME_CONST((players[team][player]).obj_id, 
				    player_team_sym, team_number_sym);

	    /* player visible */
	    (players[team][player]).visible_wme = CREATE_WME_CONST((players[team][player]).obj_id,
								   visible_symbol, no_symbol);

	    /* player distance */
	    (players[team][player]).distance_wme = CREATE_WME_VALUE((players[team][player]).obj_id,
								    distance_symbol,
								    (players[team][player]).distance_value, get_io_float_constant(-1.0));

	    /* player angle */
	    (players[team][player]).angle_wme = CREATE_WME_VALUE((players[team][player]).obj_id,
								 angle_symbol,
								 (players[team][player]).angle_value, get_io_float_constant(0.0));

	    /* player direction? XXX what is direction */
	    (players[team][player]).direction_wme = CREATE_WME_VALUE((players[team][player]).obj_id,
								     player_direction_sym_const, 
								     (players[team][player]).direction_value, get_io_float_constant(0.0));

	    /* player's absolute-position - x,y,theta */
	    (void)CREATE_WME_VALUE((players[team][player]).obj_id,
				   player_position_sym, 
				   (players[team][player]).absolute_position_obj_id,
				   get_new_io_identifier ('A'));

	    /* player's absolute-position x */
	    (players[team][player]).absolute_position_x_wme = CREATE_WME_VALUE((players[team][player]).absolute_position_obj_id,
									       player_x_symbol,
									       (players[team][player]).absolute_position_x_value,
									       get_io_float_constant(0.0));


	    /* player's absolute-position y */
	    (players[team][player]).absolute_position_y_wme = CREATE_WME_VALUE((players[team][player]).absolute_position_obj_id,
									       player_y_symbol,
									       (players[team][player]).absolute_position_y_value,
									       get_io_float_constant(0.0));


	    /* player's absolute-position theta */
	    (players[team][player]).absolute_position_theta_wme = CREATE_WME_VALUE((players[team][player]).absolute_position_obj_id,
									       player_theta_symbol,
									       (players[team][player]).absolute_position_theta_value,
									       get_io_float_constant(0.0));


				   
    
	  }	
	release_io_symbol(team_number_sym);
      }

    release_io_symbol(player_sym);
    release_io_symbol(player_number_sym);
    release_io_symbol(player_team_sym);
  }

  /* uniform, side from init msg received from server in robocup_init() */
  {
    Symbol *uniform_sym, *team_side_sym, *uniform_obj_id, *uniform_number_sym;
    Symbol *new_uniform_string_sym;
    Symbol *team_side_value_sym;
    
    char temp[10]; /* XXX uniform numbers: 2 digits max, + 1 more for '\0' */

    /* uniform */
    (void)CREATE_WME(top_input_link_id,
		     uniform_sym, get_io_sym_constant ("uniform"),
		     uniform_obj_id, get_new_io_identifier ('U'));
    
    release_io_symbol(uniform_sym);

    /* soar needs the uniform number as a string, not a number! */
    if (!sprintf(temp,"%d",MyNumber))
      {
	fprintf(stderr,"soarAppInit.c: Failed to convert the uniform number to string\n");
	exit(1);
      }

    (void)CREATE_WME(uniform_obj_id,
		     uniform_number_sym, get_io_sym_constant ("number"),
		     new_uniform_string_sym, get_io_sym_constant(temp));

    release_io_symbol(uniform_number_sym);
    release_io_symbol(new_uniform_string_sym);
    
    /* team */
    if (MySide==S_LEFT) {
      team_side_value_sym = get_io_sym_constant("l");
    } else if (S_RIGHT==MySide) {
      team_side_value_sym = get_io_sym_constant("r");
    }

    (void)CREATE_WME_ATTRIB_CONST(top_input_link_id,
				  team_side_sym, get_io_sym_constant("side"),
				  team_side_value_sym);

    release_io_symbol (team_side_sym);
    release_io_symbol (team_side_value_sym);
  }

  /* direction and power to use for home position */
  {
    Symbol *home_dir_symbol;

    (void)CREATE_WME(top_input_link_id, 
		     home_dir_symbol, get_io_sym_constant("home-direction"),
		     home_dir_obj_id, get_new_io_identifier('H'));

    release_io_symbol(home_dir_symbol);
    
    home_dir_wme = CREATE_WME_VALUE(home_dir_obj_id, angle_symbol, 
				    home_dir_value, get_io_float_constant(0.0));

    home_power_wme = CREATE_WME_VALUE(home_dir_obj_id, power_symbol, 
				      home_power_value, get_io_int_constant(0));
  }
 
  return ;
}


/*************************************************************************/
/*                  UPDATE                                               */
/*************************************************************************/


static void update_input()
{ 
  static Symbol *hear_item_sym = NULL, *hear_next_sym = NULL;
  static Symbol *hear_obj_id = NULL, *speaker_name_sym=NULL;
  static Symbol *hear_sym = NULL;
  static wme    *hear_wme = NULL;

  static Symbol *referee_hear_obj_id = NULL, *referee_name_sym=NULL;
  static Symbol *referee_hear_sym = NULL;
  static wme    *referee_hear_wme = NULL;

  static char MsgWord[BUFSIZE1];

  int i,player,team, first_word = 0;

  Symbol *current_obj_id, *spoken_sym, *next_obj_id;
  extern Bool closer_than_me (double, double, double, double);
  extern float GetKickingAngle(); /* from attacker.c? C4.5 kicking */

  /* shift memory: 
     State[PAST]     = State[PREVIOUS];
     State[PREVIOUS] = State[NOW];

     etc.
   */

  for (i=0; i< NUM_OF_STATES-1 ; i++)
    {
      memcpy(&(State[i]), &(State[i+1]), sizeof(state));
    }
    
  /*****************************
    we should first check the socket, and then
      1. all hear msgs should be immediately sent to soar
      2. only the last seen msg should be updated to soar

      Solution
      while (new msg received)
      {
        parse and put on State[NOW]
        if new hear information parsed, send to soar
      }
      
      now (after while loop),
        if there are changes in visible information
	   update them to soar
   ****************************/
      
  while (receive_and_parse())  /* state changed! */
    {
      /********************************************************************************/
      /* allow auxilary functions to supply auxilary information based on new changes */
      /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
      /********************************************************************************/

      /* interception */
      if (has_data_changed(State[NOW].Ball.SeenInfo, State[PREVIOUS].Ball.SeenInfo))
	intercept(State[NOW].Ball.SeenInfo.dist, 
		  State[NOW].Ball.SeenInfo.dir, 
		  State[NOW].Ball.SeenInfo.dist_chng, 
		  State[NOW].Ball.SeenInfo.dir_chng, 
                  State[NOW].Players[OWN][MyNumber-1].position.pos.x,
                  State[NOW].Players[OWN][MyNumber-1].position.pos.y,
                  State[NOW].Players[OWN][MyNumber-1].position.dir,
                  GlobalRole,
		  &(State[NOW].Auxilary.Intercept.angle),
		  &(State[NOW].Auxilary.Intercept.power));

      /*
	printf("****Intercept: Turn = %2.2f, Dash = %2.2f \n",
	       State[NOW].Intercept.angle, State[NOW].Intercept.power);
      */

      /* estimated current position */
      /* --> done during parsing: parse() */
      
      /* direction to opponent goal */
      /* --> done during parsing, since depends on current position */

      /* direction and power to use for home position */
      if (DIR_ERR != State[NOW].Players[OWN][MyNumber-1].position.dir)
	if (memcmp(&(State[NOW].Players[OWN][MyNumber-1].position),
		   &(State[PREVIOUS].Players[OWN][MyNumber-1].position),
		   sizeof(PosState)))
	  {
	    if (!location(State[NOW].Players[OWN][MyNumber-1].position.pos.x,
			  State[NOW].Players[OWN][MyNumber-1].position.pos.y,
			  HomePosition.pos.x, HomePosition.pos.y,
			  State[NOW].Players[OWN][MyNumber-1].position.dir,
			  &(State[NOW].Auxilary.home_direction), 
			  &(State[NOW].Auxilary.home_power)))
	      {
		fprintf(stderr,"Couldn't calculate way back home.\n");
	      }
	  }
      
      /* update smart kicking angle */
      if (memcmp(&(State[NOW].Auxilary.AnglesOfVisiblePlayers),
		 &(State[PREVIOUS].Auxilary.AnglesOfVisiblePlayers),
		 sizeof(float)*State[NOW].Auxilary.cPlayerCnt))
	State[NOW].Auxilary.smart_kick_angle = GetKickingAngle();

      /* State[NOW].Auxilary.smart_kick_angle = State[NOW].Auxilary.opponent_goal_dir; */

      /********************************************************************************/
      /* update auditory changes in the state to Soar.  Notice:  changes Only!        */
      /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
      /********************************************************************************/

      if (  (State[NOW].Msg.message[0] != '\0')
	  &&(strcmp(State[NOW].Msg.message, State[PREVIOUS].Msg.message)))
	{
	  /* this is probably the only place where an initialization is
	     appropriate here rather than in initialize_top_state().  This
	     is because we shouldn't create an empty hear id (so we shouldn't
	     just create it and wait for an actual msg to fill the contents)
	   */
	      
	  switch (State[NOW].Msg.sender) {
   	    case ST_Referee:
	      /* printf("received msg from referee: %s\n",State[NOW].Msg.message); */
	      if (!referee_hear_wme)
		{
		  referee_hear_sym = get_io_sym_constant ("referee-hear");
		  
		  referee_hear_wme = CREATE_WME_VALUE(top_input_link_id, 
						      referee_hear_sym, 
						      referee_hear_obj_id, get_new_io_identifier ('R'));

		  /* these might have been initialized before */
		  if (!hear_item_sym)
		    hear_item_sym = get_io_sym_constant ("item");
		  if (!hear_next_sym)
		    hear_next_sym = get_io_sym_constant ("next");
		}		  
	      else
		{
		  /* we want to create a new hear id every time to signal 
		     that a new msg has been received. 
		     */
		  
		  UPDATE_WME(referee_hear_wme, top_input_link_id, referee_hear_sym,
			     referee_hear_obj_id, get_new_io_identifier ('R'));
		}

	      current_obj_id = referee_hear_obj_id;
	      first_word = 0; 
	      break;

	  case ST_Self: /* intentionally letting it slip into the next one */
	  case ST_Player:
	    if (!(hear_wme))
	      {
		hear_sym = get_io_sym_constant ("hear");
		
		hear_wme = CREATE_WME_VALUE(top_input_link_id, 
					    hear_sym, 
					    hear_obj_id, get_new_io_identifier ('h'));
		
		/* these might have been initialized before */
		if (!hear_item_sym)
		  hear_item_sym = get_io_sym_constant ("item");
		if (!hear_next_sym)
		  hear_next_sym = get_io_sym_constant ("next");
	      }		  
	    else
	      {
		/* we want to create a new hear id every time to signal 
		   that a new msg has been received. 
		   */
		
		UPDATE_WME(hear_wme, top_input_link_id, hear_sym,
			   hear_obj_id, get_new_io_identifier ('h'));
	      }

	    current_obj_id = hear_obj_id;
	    first_word = 1; /* skip teamname in word #0 */

	    break;
	  }
	  
	  /* note we don't have to save the pointers to the following
	     wmes and release them, since Soar garbage collection is going to
	     do this for us:  once we discard the hear_wme, which points to
	     these, Soar will also discard them */

	  /* now get word by word from msg, and put it in wmes 
	     (Notice for our own msgs, we skip the first word which
	      is the Teamname) */
	  for (i=first_word; get_word(State[NOW].Msg.message,i,MsgWord); i++)
	    {
	      /* create word under item */
	      (void)CREATE_WME_VALUE(current_obj_id, hear_item_sym, 
				     spoken_sym, get_io_sym_constant (MsgWord));
	      release_io_symbol(spoken_sym);

	      /* create pointer to next word */
	      (void)CREATE_WME_VALUE(current_obj_id, hear_next_sym,
				     next_obj_id, get_new_io_identifier ('n'));

	      current_obj_id = next_obj_id;
	    }
	  /* end nicely last item as NULL */
	  /*
	    (void)CREATE_WME_VALUE(current_obj_id, hear_item_sym,
	                           spoken_sym,get_io_sym_constant("NULL"));
	  */
	}
    }

  /* check whether it is OK to now take an action on the output */

  /* commented out by Milind 3 Aug 97

  check_action_timer();

  */

  /* now update any new information based on visual sensory */
  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

  /* server time */
  if (State[NOW].time != State[PREVIOUS].time)
    {
      UPDATE_WME(value_wme,              /* wme to be updated */
		 time_obj_id, value_sym, /* root and attribute */
		 new_int_sym, get_io_int_constant(State[NOW].time)); /* value var. and new value */
    }

  /* recommendation for kicking angle */
  if (   State[NOW].Auxilary.opponent_goal_dir 
      != State[PREVIOUS].Auxilary.opponent_goal_dir)
    {
      UPDATE_WME(kick_angle_wme,
		 top_input_link_id, kick_angle_symbol,
		 kick_angle_value, get_io_float_constant(State[NOW].Auxilary.opponent_goal_dir));
    }

  /* recommendation for smart kicking angle */
  if (   State[NOW].Auxilary.smart_kick_angle
      != State[PREVIOUS].Auxilary.smart_kick_angle)
    {
      UPDATE_WME(smart_kick_wme, top_input_link_id, smart_kick_symbol, 
		 smart_kick_value, 
		 get_io_float_constant(State[NOW].Auxilary.smart_kick_angle));
    }

  /* recommendation for clearing the ball angle */
  if (   State[NOW].Auxilary.clear_ball_angle
      != State[PREVIOUS].Auxilary.clear_ball_angle)
    {
      UPDATE_WME(clear_ball_wme, top_input_link_id, clear_ball_symbol, 
		 clear_ball_value, 
		 get_io_float_constant(State[NOW].Auxilary.clear_ball_angle));
    }

  /* ball */
  if (  (visible(State[NOW].Ball))
	&&(!visible(State[PREVIOUS].Ball)))
    {
      /* ball is now visible, was previously not */
      
      /* update ball visibility wme */
      UPDATE_WME_CONST(ball_visible_wme, 
		       ball_obj_id, visible_symbol, yes_symbol);
    }
  else if (  (!visible(State[NOW].Ball))
	     &&(visible(State[PREVIOUS].Ball)))
    {
      /* ball is no longer visible.  Do not change last known data, though */
      
      /* update ball visibility wme */
      UPDATE_WME_CONST(ball_visible_wme, ball_obj_id, visible_symbol,
		       no_symbol);
    }
	
  if (has_data_changed(State[NOW].Ball.SeenInfo, State[PREVIOUS].Ball.SeenInfo))
    {
      /* update ball distance wme */
      UPDATE_WME(new_ball_dist_wme, ball_obj_id, distance_symbol,
		 ball_distance_value_sym, get_io_float_constant(State[NOW].Ball.SeenInfo.dist));
      
      /* update ball angle wme */
      UPDATE_WME(new_ball_angle_wme, ball_obj_id, angle_symbol,
		 ball_angle_value_sym, get_io_float_constant(State[NOW].Ball.SeenInfo.dir));
    }
 

  /* update interception wmes */
  if (memcmp(&(State[NOW].Auxilary.Intercept), 
	     &(State[PREVIOUS].Auxilary.Intercept),
	     sizeof(InterceptInfo)))
    {
      UPDATE_WME(intercept_angle_wme, intercept_obj_id, angle_symbol,
		 intercept_angle_value_sym, 
		 get_io_float_constant((float)State[NOW].Auxilary.Intercept.angle));
      
      UPDATE_WME(intercept_power_wme, intercept_obj_id, power_symbol,
		 intercept_power_value_sym, 
		 get_io_int_constant((int)(State[NOW].Auxilary.Intercept.power)));
    }


  /* flags */
  for (i=0; i<NUM_FLAGS; i++)
    {
      if (  (visible(State[NOW].Flags[i]))
	    &&(!visible(State[PREVIOUS].Flags[i])))
	{
	  /* flag is now visible, it wasn't before */

	  /* update visibility wme */
	  UPDATE_WME_CONST(flags[i].visible_wme, flags[i].obj_id, visible_symbol, yes_symbol);
	}
      else if (  (!visible(State[NOW].Flags[i]))
		 &&(visible(State[PREVIOUS].Flags[i])))  
	{
	  /* flag is no longer visible */
	  
	  /* update visibility wme */
	  UPDATE_WME_CONST(flags[i].visible_wme, flags[i].obj_id, visible_symbol, no_symbol);
	}

      if (has_data_changed(State[NOW].Flags[i].SeenInfo, State[PREVIOUS].Flags[i].SeenInfo))
	{
	  /* update distance wme */
	  UPDATE_WME(flags[i].distance_wme, flags[i].obj_id, distance_symbol,
		     flags[i].distance_value, get_io_float_constant(State[NOW].Flags[i].SeenInfo.dist));
	  
	  /* update angle wme */
	  UPDATE_WME(flags[i].angle_wme, flags[i].obj_id, angle_symbol,
		     flags[i].angle_value, get_io_float_constant(State[NOW].Flags[i].SeenInfo.dir));
	}
    }

  /* players and teams */
  State[NOW].Auxilary.num_friendlies_closer = 0;
  State[NOW].Auxilary.num_enemies_closer    = 0;

  for (team = 0; team<TEAMS; team++)
    {
      for (player = 0; player<MAX_PLAYERS; player++) 
	{
	  if (  (visible(State[NOW].Players[team][player]))
		&&(!visible(State[PREVIOUS].Players[team][player])))
	    {
	      /** player is visible now, wasn't before */
	      
	      /* update visibility wme */
	      UPDATE_WME_CONST(players[team][player].visible_wme, players[team][player].obj_id, 
			       visible_symbol, yes_symbol);
	    }
	  else if (  (!visible(State[NOW].Players[team][player]))
		   &&(visible(State[PREVIOUS].Players[team][player])))
	    {
	      /** player is not visible now, was before */
	      
	      /* update visibility wme */
	      UPDATE_WME_CONST(players[team][player].visible_wme, players[team][player].obj_id, 
			       visible_symbol, no_symbol);
	    }

	  if (has_data_changed(State[NOW].Players[team][player].SeenInfo,
			       State[PREVIOUS].Players[team][player].SeenInfo))
	    {
	      /* player date has changed */
	      
	      /* update player distance */
	      UPDATE_WME(players[team][player].distance_wme,players[team][player].obj_id,
			 distance_symbol, players[team][player].distance_value,
			 get_io_float_constant(State[NOW].Players[team][player].SeenInfo.dist));

	      /* update player angle */
	      UPDATE_WME(players[team][player].angle_wme,players[team][player].obj_id,
			 angle_symbol, players[team][player].angle_value,
			 get_io_float_constant(State[NOW].Players[team][player].SeenInfo.dir));

	      /* update player direction */
	      UPDATE_WME(players[team][player].direction_wme,players[team][player].obj_id,
			 player_direction_sym_const, players[team][player].direction_value,
			 get_io_float_constant(State[NOW].Players[team][player].SeenInfo.dir_chng));
	    } /* player data changed */

	  /* do this only for our own position */
	  if (  (MyNumber-1 == player)
	      &&(OWN == team))
	    {
	      if (memcmp(&(State[NOW].Players[team][player].position),
			 &(State[PREVIOUS].Players[team][player].position),
			 sizeof(PosState)))
		{
		  
		  /* player's position data has changed  */
		  
		  /* update x absolute position  */
		  UPDATE_WME(players[team][player].absolute_position_x_wme,
			     players[team][player].absolute_position_obj_id,
			     player_x_symbol, players[team][player].absolute_position_x_value,
			     get_io_float_constant((float)State[NOW].Players[team][player].position.pos.x));
		  
		  /* update y absolute position  */
		  UPDATE_WME(players[team][player].absolute_position_y_wme,
			     players[team][player].absolute_position_obj_id,
			     player_y_symbol, players[team][player].absolute_position_y_value,
			     get_io_float_constant((float)State[NOW].Players[team][player].position.pos.y));
		  
		  /* update theta absolute position  */
		  UPDATE_WME(players[team][player].absolute_position_theta_wme,
			     players[team][player].absolute_position_obj_id,
			     player_theta_symbol, players[team][player].absolute_position_theta_value,
			     get_io_float_constant((float)State[NOW].Players[team][player].position.dir));
		}
	    }

	  /* if greater than ball close distance test angle and distance 
	     else just test to the player */
	  if ((visible(State[NOW].Ball)) && 
	      (visible(State[NOW].Players[team][player])) &&
	      (closer_than_me(State[NOW].Players[team][player].SeenInfo.dir,
			     State[NOW].Players[team][player].SeenInfo.dist,
			     State[NOW].Ball.SeenInfo.dir,
			     State[NOW].Ball.SeenInfo.dist)))
	    {
	      if (OWN==team) /* my team */
		{
		  (State[NOW].Auxilary.num_friendlies_closer)++;
		}
	      else /* team == OPPONENT */
		{
		  (State[NOW].Auxilary.num_enemies_closer)++;
		}
	    }
		  
	} /* MAX_PLAYERS */
		  
    }   /* TEAMS */
	

  /* update us_close_to_ball, etc. */
  if (  State[NOW].Auxilary.num_friendlies_closer
      !=State[PREVIOUS].Auxilary.num_friendlies_closer)
    {

      /* XXX GalK following if works because we maintain the soar.visible
	 between percepts.  So that PREVIOUS == NEXT if we didn't get a new
	 see msg.  But this is really a bug, since between percepts the ball
	 isn't visible, now is it!? */

      /* update us_close */
      if (  (State[NOW].Auxilary.num_friendlies_closer > 0) 
	  &&(0 == State[PREVIOUS].Auxilary.num_friendlies_closer))
	{
	  UPDATE_WME_CONST(new_ball_us_close_wme, ball_obj_id,
			   ball_us_close_sym,yes_symbol);
	}
      else if (  (State[PREVIOUS].Auxilary.num_friendlies_closer > 0)
	       &&(0 == State[NOW].Auxilary.num_friendlies_closer))
	{
	  UPDATE_WME_CONST(new_ball_us_close_wme, ball_obj_id,
			   ball_us_close_sym,no_symbol);
	}
	
      /* update count */
      UPDATE_WME(close_count[OWN].close_wme, ball_obj_id,
		 close_count[OWN].close_symbol, close_count[OWN].close_value,
		 get_io_int_constant(State[NOW].Auxilary.num_friendlies_closer));
    }

  if (  State[NOW].Auxilary.num_enemies_closer
      !=State[PREVIOUS].Auxilary.num_enemies_closer)
    {
      /* XXX GalK following if works because we maintain the soar.visible
	 between percepts.  So that PREVIOUS == NEXT if we didn't get a new
	 see msg.  But this is really a bug, since between percepts the ball
	 isn't visible, now is it!? */

      /* update other_close */
      if (  (State[NOW].Auxilary.num_enemies_closer > 0) 
	  &&(0 == State[PREVIOUS].Auxilary.num_enemies_closer))
	{
	  UPDATE_WME_CONST(new_ball_other_close_wme, ball_obj_id,
			   ball_other_close_sym,yes_symbol);
	}
      else if (  (State[PREVIOUS].Auxilary.num_enemies_closer > 0)
	       &&(0 == State[NOW].Auxilary.num_enemies_closer))
	{
	  UPDATE_WME_CONST(new_ball_other_close_wme, ball_obj_id,
			   ball_other_close_sym,no_symbol);
	}
	
      /* update count */
      UPDATE_WME(close_count[OPPONENT].close_wme, ball_obj_id,
		 close_count[OPPONENT].close_symbol, close_count[OPPONENT].close_value,
		 get_io_int_constant(State[NOW].Auxilary.num_enemies_closer));
    }

  /* update direction and power to use for home position */
  if (  State[NOW].Auxilary.home_direction
      !=State[PREVIOUS].Auxilary.home_direction)
    {
      UPDATE_WME(home_dir_wme, home_dir_obj_id,
		 angle_symbol, home_dir_value,
		 get_io_float_constant(State[NOW].Auxilary.home_direction));
    }

  if (  State[NOW].Auxilary.home_power
      !=State[PREVIOUS].Auxilary.home_power)
    {
      UPDATE_WME(home_power_wme, home_dir_obj_id,
		 power_symbol, home_power_value,
		 get_io_int_constant(State[NOW].Auxilary.home_power));
    }

  /* goals */
  for (i=0; i<TEAMS; i++)
    {
      if (  (visible(State[NOW].Goals[i]))
	  &&(!visible(State[PREVIOUS].Goals[i])))
	{
	  UPDATE_WME_CONST(goals[i].visible_wme,
			   goals[i].obj_id,visible_symbol,yes_symbol);
	}
      else if (  (!visible(State[NOW].Goals[i]))
		 &&(visible(State[PREVIOUS].Goals[i])))
	{
	  UPDATE_WME_CONST(goals[i].visible_wme,
			   goals[i].obj_id,visible_symbol,no_symbol);
	}

      if  (has_data_changed(State[NOW].Goals[i].SeenInfo,
			    State[PREVIOUS].Goals[i].SeenInfo))
	{
	  /* update goal distance */
	  UPDATE_WME(goals[i].distance_wme, 
		     goals[i].obj_id, distance_symbol,
		     goals[i].distance_value, get_io_float_constant(State[NOW].Goals[i].SeenInfo.dist));
	      
	  /* update goal angle */
	  UPDATE_WME(goals[i].angle_wme,
		     goals[i].obj_id, angle_symbol,
		     goals[i].angle_value, get_io_float_constant(State[NOW].Goals[i].SeenInfo.dir));
	}
    }

  return ;
  
} /* input function ends */



void my_input_function (soar_callback_agent the_agent, 
			soar_callback_data data,
			soar_call_data call_data)

{
  agent * a = (agent *) the_agent;
  int mode = (int) call_data;

  switch (mode) {
  
  case TOP_STATE_JUST_CREATED:
    /* Initialize & internalize objects. */
    initialize_top_state(a);

    break;

  case TOP_STATE_JUST_REMOVED:
    /* Clean up. */
    release_io_symbol (time_obj_id);
    release_io_symbol (value_sym);
    release_io_symbol (new_int_sym);
    release_io_symbol (top_input_link_id);
    /* got to release everything here */

    break;

  case NORMAL_INPUT_CYCLE:
    /* Standard mode of operation. */
    update_input();

    break;

  }

  return ;
}

