/*
 *----------------------------------------------------------------------
 *
 * PLEASE NOTE!  Only functions implementing commands should appear
 * in this file.  All supporting functions should be placed in 
 * command_utils.c.
 *
 *----------------------------------------------------------------------
 */

#include <soar.h>

#include <soarCommands.h>
#include <soarCommandUtils.h>
#include <soarScheduler.h>

/*
 *----------------------------------------------------------------------
 *
 * IOCmd --
 *
 *      This is the command procedure for the "io" command, 
 *      which manages the attachment of scripts to handle Soar I/O.
 *
 * Syntax:  io <add-specification> | 
 *             <removal-specification> |
 *             <list-specification> |
 *
 *          <add-specification>     := -add [-input | -output ] script [id]
 *          <removal-specification> := -delete [-input | -output] id
 *          <list-specification>    := -list [-input | -output]
 *
 * Results:
 *      Returns a standard Tcl completion code. 
 *
 * Side effects:
 *      Adds and/or removes indicated I/O procedure from the system.
 *      This command may also print information about defined I/O
 *      procedures.  Returns the name of the new I/O procedure
 *      if one is created.
 *
 *----------------------------------------------------------------------
 */

static int io_proc_counter = 1;

int IOCmd (ClientData clientData, 
	   Tcl_Interp * interp,
	   int argc, char *argv[])
{
  static char * too_few_args_string = "Too few arguments, should be: io [-add [-input | -output] script [id]] | [-delete [-input|-output] id] | [-list [-input|-output]";
  static char * too_many_args_string = "Too many arguments, should be: io [-add [-input|-output] script [id]] | [-delete [-input|-output] id] | [-list [-input|-output]";
  char * io_id;
  char   buff[10];          /* What size is good here? */

  if (argc == 1)
    {
      interp->result = too_few_args_string;
      return TCL_ERROR;
    }
  if (string_match_up_to(argv[1], "-add", 2))
    {
      if (argc < 4)
	{
	  interp->result = too_few_args_string;
	  return TCL_ERROR;
	}

      if (argc > 5)
	{
	  interp->result = too_many_args_string;
	  return TCL_ERROR;
	}

      if (argc == 5)
	{
	  io_id = argv[4];
	}
      else
	{
	  sprintf(buff, "m%d", io_proc_counter++);
	  io_id = buff;
	}

      {
	SOAR_CALLBACK_TYPE ct;
	soar_callback_fn fn;
	
	if (string_match_up_to(argv[2], "-input", 2))
	  {
	    add_input_function((agent *) clientData, 
			       soar_input_callback_to_tcl,
			       (soar_callback_data) savestring(argv[3]), 
			       soar_callback_data_free_string,
			       (soar_callback_id) io_id);
	  }
	else if (string_match_up_to(argv[2], "-output", 2))
	  {
	    add_output_function((agent *) clientData, 
				soar_output_callback_to_tcl,
				(soar_callback_data) savestring(argv[3]), 
				soar_callback_data_free_string,
				(soar_callback_id) io_id);
	  }
	else
	  {
	    sprintf(interp->result,
		    "%s: Unrecognized IO type: %s %s",
		    argv[0], argv[1], argv[2]);
	    return TCL_ERROR;
	  }

	interp->result = io_id;
	return TCL_OK;
      }
    }
  else if (string_match_up_to(argv[1], "-delete", 2))
    {
      switch (argc) {
      case 2:
      case 3:	  
	interp->result = too_few_args_string;
	return TCL_ERROR;
      case 4:	  /* Delete single callback for given event */
	  {
	    if (string_match_up_to(argv[2], "-input", 2))
	      {
		remove_input_function((agent *) clientData, argv[3]);
	      }
	    else if (string_match_up_to(argv[2], "-output", 2))
	      {
		remove_output_function((agent *) clientData, argv[3]);
	      }
	    else
	      {
		sprintf(interp->result,
			"Attempt to delete unrecognized io type: %s",
			argv[2]);
		return TCL_ERROR;
	      }
	  }
	  break;
      default:
	interp->result = too_many_args_string;
	return TCL_ERROR;
      }
    }
  else if (string_match_up_to(argv[1], "-list", 2))
    {
      switch (argc) {
      case 2:	  
	interp->result = too_few_args_string;
	return TCL_ERROR;
      case 3:	  
	{
	  SOAR_CALLBACK_TYPE ct;
	  
	  if (string_match_up_to(argv[2], "-input", 2))
	    {
	      ct = INPUT_PHASE_CALLBACK;
	    }
	  else if (string_match_up_to(argv[2], "-output", 2))
	    {
	      ct = OUTPUT_PHASE_CALLBACK;
	    }
	  else
	    {
	      sprintf(interp->result,
		      "Attempt to list unrecognized io type: %s",
		      argv[2]);
	      return TCL_ERROR;
	    }

	  soar_push_callback((soar_callback_agent) clientData,
			     PRINT_CALLBACK,
			     (soar_callback_fn) Soar_AppendResult, 
			     (soar_callback_data) NULL,
			     (soar_callback_free_fn) NULL);
	  soar_list_all_callbacks_for_event((agent *) clientData, ct);
	  soar_pop_callback((soar_callback_agent) clientData, 
			    PRINT_CALLBACK);
	}
	break;
      default:
	interp->result = too_many_args_string;
	return TCL_ERROR;
      }

    }
  else
    {
      sprintf(interp->result, 
	      "Unrecognized option to io command: %s", argv[1]);
      return TCL_ERROR;
    }
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AddWmeCmd --
 *
 *      This is the command procedure for the "add-wme" command, 
 *      which adds an element to working memory.
 *
 *      This command surgically modifies Soar's working memory.  
 *      Add-wme adds a new wme with the given id, attribute, value, 
 *      and optional acceptable preference.  The given id must be an 
 *      existing identifier.  If '*' is given in place of the 
 *      attribute or value, Soar creates a new identifier (gensym)
 *      for that field.  
 *
 *      WARNING: this command is inherently unstable and may have 
 *      weird side effects (possibly even including system crashes).  
 *      For example, the chunker can't backtrace through wmes created 
 *      via add-wme.  Removing input wmes or context/impasse wmes may 
 *      have unexpected side effects.  You've been warned.
 *
 *      See also:  remove-wme
 *
 * Syntax:  add-wme id [^] { attribute | '*'} { value | '*' } [+]
 *
 * Results:
 *      Returns a standard Tcl completion code and the timetag of the
 *      created wme (which can be passed to remove-wme).
 *
 * Side effects:
 *      Adds the given wme to working memory.
 *
 *----------------------------------------------------------------------
 */

int AddWmeCmd (ClientData clientData, 
	       Tcl_Interp * interp,
	       int argc, char *argv[])
{
  static char * too_few_args  = "Too few arguments.\nUsage: add-wme id [^] { attribute | '*'} { value | '*' } [+]";
  static char * too_many_args = "Too many arguments.\nUsage: add-wme id [^] { attribute | '*'} { value | '*' } [+]";

  Symbol *id, *attr, *value;       /* Symbols used to create the WME */
  bool acceptable_preference;      /* Acceptable preference seen?    */
  wme *w;                          /* The created WME                */
  int next_arg;                    /* Index into argv used to point  */
                                   /* to next command argument       */
  if (argc < 4)
    {
      interp->result = too_few_args;
      return TCL_ERROR;
    }

  /* --- get id --- */

  if (read_id_or_context_var_from_string(argv[1], &id) 
      == TCL_ERROR)
    {
      sprintf(interp->result,
	      "Known id or context var expected in add-wme command, not %s.",
	      argv[1]);
      return TCL_ERROR;
    }

  /* --- get optional '^', if present --- */

  /* With Tcl parsing, the ^ could be separate or a prefix of   */
  /* the attribute lexeme.                                      */

  /* Unlike non-Tcl version, don't err if missing.              */
  
  if (string_match("^", argv[2]))
    {
      next_arg = 3;
    }
  else
    {
      next_arg = 2;
      if (*argv[2] == '^')
	{
	  argv[2]++;                 /* Skip leading ^ */
	}
    }
      
  /* --- get attribute or '*' --- */

  if (string_match("*", argv[next_arg]))
    {
      attr = make_new_identifier ('I', id->id.level);
    } 
  else 
    {
      get_lexeme_from_string(argv[next_arg]);

      switch (current_agent(lexeme).type) 
	{
	case SYM_CONSTANT_LEXEME:
	  attr = make_sym_constant (current_agent(lexeme).string); 
	  break;
	case INT_CONSTANT_LEXEME:
	  attr = make_int_constant (current_agent(lexeme).int_val); 
	  break;
	case FLOAT_CONSTANT_LEXEME:
	  attr = make_float_constant (current_agent(lexeme).float_val); 
	  break;
	case IDENTIFIER_LEXEME:
	case VARIABLE_LEXEME:      
	  attr = read_identifier_or_context_variable();
	  if (!attr) 
	    {
	      return TCL_ERROR;
	    }
	  symbol_add_ref (attr);
	  break;
	default:
	  sprintf (interp->result,
		   "Expected constant, identifier, or '*' for attribute, not %s.",
		   argv[next_arg]);
	  return TCL_ERROR;
	}
    }

  /* --- get value or '*' --- */

  next_arg++;
  if (next_arg >= argc)
    {
      symbol_remove_ref (attr); 
      interp->result = too_few_args;
      return TCL_ERROR;
    }

  if (string_match("*", argv[next_arg]))
    {
      value = make_new_identifier ('I', id->id.level);
    } 
  else 
    {
      get_lexeme_from_string(argv[next_arg]);
      switch (current_agent(lexeme).type) 
	{
	case SYM_CONSTANT_LEXEME:
	  value = make_sym_constant (current_agent(lexeme).string); 
	  break;
	case INT_CONSTANT_LEXEME:
	  value = make_int_constant (current_agent(lexeme).int_val); 
	  break;
	case FLOAT_CONSTANT_LEXEME:
	  value = make_float_constant (current_agent(lexeme).float_val); 
	  break;
	case IDENTIFIER_LEXEME:
	case VARIABLE_LEXEME:
	  value = read_identifier_or_context_variable();
	  if (!value) 
	    { 
	      symbol_remove_ref (attr); 
	      return TCL_ERROR; 
	    }
	  symbol_add_ref (value);
	  break;
	default:
	  sprintf (interp->result,
		   "Expected constant, identifier, or '*' for value, not %s.",
		   argv[next_arg]);
	  symbol_remove_ref (attr);
	  return TCL_ERROR;
	}
    }

  /* --- get optional acceptable preference indicator --- */

  next_arg++;

  if (string_match("+", argv[next_arg]))
    { 
      acceptable_preference = TRUE; 
      next_arg++;
    }
  else if (next_arg != argc)      /* An arg is present and != "+" */
    {
      symbol_remove_ref (attr); 
      symbol_remove_ref (value); 
      sprintf (interp->result,
	       "Only '+' allowed as optional preference, not %s.",
	       argv[next_arg]);
      return TCL_ERROR;      
    }
  else
    {
      acceptable_preference = FALSE;
    }

  if (next_arg < argc)                     /* Not at end of argv */
    {
      symbol_remove_ref (attr); 
      symbol_remove_ref (value); 
      interp->result = too_many_args;
      return TCL_ERROR;      
    }

  /* --- now create and add the wme --- */

  w = make_wme (id, attr, value, acceptable_preference);
  symbol_remove_ref (attr);
  symbol_remove_ref (value);
  insert_at_head_of_dll (id->id.input_wmes, w, next, prev);
  add_wme_to_wm (w);
  do_buffered_wm_and_ownership_changes();

  /* --- return wme timetag and final wme as result */

  soar_push_callback((soar_callback_agent) clientData, 
		     PRINT_CALLBACK,
		     (soar_callback_fn) Soar_AppendResult, 
		     (soar_callback_data) NULL,
		     (soar_callback_free_fn) NULL);

  print_wme_for_tcl(w);
  soar_pop_callback((soar_callback_agent) clientData, PRINT_CALLBACK);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AgentGoCmd --
 *
 *	This is the command procedure for the "agent-go" command, 
 *	which defines the default "go" settings for the agent.
 *
 *      This command operates exactly the same as the go command, 
 *      EXCEPT that this only defines the go settings for a agent 
 *      when it is selected to run by the multi-agent scheduler.  
 *      Hence, the go settings are defined for subsequent runs, but 
 *      no runs are made when the command is read
 *      See also:  go
 *
 * Syntax:  agent-go agent-name [integer | 'forever'] [type]
 *          type ::= p | e | d | g | ps | s | o | context-variable
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Redefines the default settings for the "go" command.
 *
 *----------------------------------------------------------------------
 */

int AgentGoCmd (ClientData clientData, 
		Tcl_Interp * interp,
		int argc, char *argv[])
{
  cons  * c;
  agent * the_agent;

  for (c = all_soar_agents; c != NIL; c = c->rest) 
    {
      the_agent = (agent *) c->first;

      if (string_match(the_agent->name, argv[1]))
	{
	  int parse_result;
	  long               go_number;
	  enum go_type_enum  go_type;
	  Symbol *           go_slot_attr;
	  goal_stack_level   go_slot_level;
 
	  go_number     = the_agent->go_number;
	  go_type       = the_agent->go_type;
	  go_slot_attr  = the_agent->go_slot_attr;
	  go_slot_level = the_agent->go_slot_level;

	  parse_result = parse_go_command(interp, argc-1, argv+1,
					  &go_number, &go_type,
					  &go_slot_attr, &go_slot_level);
	  if (parse_result == TCL_OK) 
	    {
	      the_agent->go_number     = go_number;
	      the_agent->go_type       = go_type;
	      the_agent->go_slot_attr  = go_slot_attr;
	      the_agent->go_slot_level = go_slot_level;
      	    }

	  return parse_result;
	}
    }

  sprintf(interp->result,
	  "Agent %s is not known.  Ignoring command.", 
	  argv[1]);
  return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * AttributePreferencesModeCmd --
 *
 *	This is the command procedure for the "attribute-preferences-mode"
 *      command, which controls the handling of preferences (other than 
 *      acceptable and reject preferences) for non-context slots.  It takes
 *      a single numeric argument: 
 *          0  --> Handle them the normal (Soar 6) way.
 *          1  --> Handle them the normal (Soar 6) way, but print a warning
 *                 message whenever a preference other than + or - is created
 *                 for a non-context slot.
 *          2  --> Whenever a preferences other than + or - is created for a 
 *                 non-context slot, print an error message and ignore
 *                 (discard) that preference.  For non-context slots, the set
 *                 of values installed in working memory is always equal to
 *                 the set of acceptable values minus the set of rejected
 *                 values.
 *
 * Syntax:  attribute-preferences-mode {0 | 1 | 2}
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Sets current_agent(attribute_preferences_mode).
 *
 *----------------------------------------------------------------------
 */

int AttributePreferencesModeCmd (ClientData clientData, 
				 Tcl_Interp * interp,
				 int argc, char *argv[])
{
  if (argc > 2)
    {
      interp->result =
        "Too many arguments.\nUsage: attribute-preferences-mode 0|1|2";
      return TCL_ERROR;
    }

  if (argc == 2) {
    if (! strcmp(argv[1], "0")) {
      current_agent(attribute_preferences_mode) = 0;
    } else if (! strcmp(argv[1], "1")) {
      current_agent(attribute_preferences_mode) = 1;
    } else if (! strcmp(argv[1], "2")) {
      current_agent(attribute_preferences_mode) = 2;
    } else {
      sprintf(interp->result,
	      "Unrecognized argument to %s: %s.  Integer 0, 1, or 2 expected.",
	      argv[0], argv[1]);
      return TCL_ERROR;
    }
  }

  sprintf(interp->result, "%d", current_agent(attribute_preferences_mode));
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ChunkFreeProblemSpacesCmd --
 *
 *      This is the command procedure for the "chunk-free-problem-spaces" 
 *      command, which manages the chunk-free Soar problem spaces.
 *
 *      With no arguments, this command prints the current list of 
 *      problem spaces declared chunk-free.  With arguments, it adds or 
 *      removes a given problem space from this list.  No chunks will be 
 *      built in a problem space if that space is declared chunk-free.
 *
 *      See also:  learn, chunky-problem-spaces
 *
 * Syntax:  chunk-free-problem-spaces [[-add space-name] | 
 *                                     [-remove space-name]]
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Modifies the chunk-free problem space list.  Prints the
 *      chunk-free problem space list.
 *
 *----------------------------------------------------------------------
 */

int CFPSCmd (ClientData clientData, 
	     Tcl_Interp * interp,
	     int argc, char *argv[])
{
  if (argc == 1)
    {
      cons * c;
      char buff[1024];

      for (c = current_agent(chunk_free_problem_spaces); 
	   c != NIL; 
	   c = c->rest)
	{
	  symbol_to_string((Symbol *) (c->first), TRUE, buff);
	  Tcl_AppendElement(interp, buff);
	}
      return TCL_OK;
    }
  
  if (argc == 2)
    {
      interp->result = "Too few args, should be: chunk-free-problem-spaces [[-add space-name] | [-remove space-name]]";
      return TCL_ERROR;
    }

  if (argc > 3)
    {
      interp->result = "Too many args, should be: chunk-free-problem-spaces [[-add space-name] | [-remove space-name]]";
      return TCL_ERROR;
    }

  if (string_match_up_to(argv[1], "-add", 2))
    {
      Symbol * sym; 

      sym = make_sym_constant (argv[2]);
      if (! member_of_list (sym, current_agent(chunk_free_problem_spaces))) {
        symbol_add_ref (sym);
        push (sym, current_agent(chunk_free_problem_spaces));
      }
      symbol_remove_ref (sym);
    }
  else if (string_match_up_to(argv[1], "-remove", 2))
    {
      Symbol * sym;
      list * extracted_stuff;

      sym = find_sym_constant (argv[2]);
      if (sym) 
	{
	  space_to_remove_from_cfps = sym;
	  extracted_stuff = extract_list_elements (&current_agent(chunk_free_problem_spaces),
						   cfps_removal_test_function);
	  deallocate_symbol_list_removing_references (extracted_stuff);
	  return TCL_OK;
	} 
      else
	{
	  sprintf(interp->result,
		  "Attempt to remove unrecognized chunk-free problem space name: %s",
		  argv[2]);
	  return TCL_ERROR;
	}
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized option to chunk-free-problem-spaces command: %s",
	      argv[1]);
      return TCL_ERROR;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ChunkyProblemSpacesCmd --
 *
 *      This is the command procedure for the "chunky-problem-spaces" 
 *      command, which manages the chunky Soar problem spaces.
 *
 *      With no arguments, this command prints the current list of 
 *      problem spaces declared chunky.  With arguments, it adds or 
 *      removes a given problem space from this list.  Chunks will be 
 *      built in a chunky problem space if that space is declared 
 *      chunky and learn is set to "specify".
 *
 *      See also:  learn, chunk-free-problem-spaces
 *
 * Syntax:  chunk-free-problem-spaces [[-add space-name] | 
 *                                     [-remove space-name]]
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Modifies the chunky problem space list.  Prints the
 *      chunky problem space list.
 *
 *----------------------------------------------------------------------
 */

int CPSCmd (ClientData clientData, 
	    Tcl_Interp * interp,
	    int argc, char *argv[])
{
  if (argc == 1)
    {
      cons * c;
      char buff[1024];

      for (c = current_agent(chunky_problem_spaces); 
	   c != NIL; 
	   c = c->rest)
	{
	  symbol_to_string((Symbol *) (c->first), TRUE, buff);
	  Tcl_AppendElement(interp, buff);
	}
      return TCL_OK;
    }
  
  if (argc == 2)
    {
      interp->result = "Too few args, should be: chunky-problem-spaces [[-add space-name] | [-remove space-name]]";
      return TCL_ERROR;
    }

  if (argc > 3)
    {
      interp->result = "Too many args, should be: chunky-problem-spaces [[-add space-name] | [-remove space-name]]";
      return TCL_ERROR;
    }

  if (string_match_up_to(argv[1], "-add", 2))
    {
      Symbol * sym; 

      sym = make_sym_constant (argv[2]);
      if (! member_of_list (sym, current_agent(chunky_problem_spaces))) {
        symbol_add_ref (sym);
        push (sym, current_agent(chunky_problem_spaces));
      }
      symbol_remove_ref (sym);
    }
  else if (string_match_up_to(argv[1], "-remove", 2))
    {
      Symbol * sym;
      list * extracted_stuff;

      sym = find_sym_constant (argv[2]);
      if (sym) 
	{
	  space_to_remove_from_cfps = sym;
	  extracted_stuff = extract_list_elements (&current_agent(chunky_problem_spaces),
						   cfps_removal_test_function);
	  deallocate_symbol_list_removing_references (extracted_stuff);
	  return TCL_OK;
	} 
      else
	{
	  sprintf(interp->result,
		  "Attempt to remove unrecognized chunky problem space name: %s",
		  argv[2]);
	  return TCL_ERROR;
	}
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized option to chunky-problem-spaces command: %s",
	      argv[1]);
      return TCL_ERROR;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DefWmeDepthCmd --
 *
 *      This is the command procedure for the "default-wme-depth" 
 *      command, which sets/prints the default print depth.
 *
 *      With no arguments, this command prints the current default 
 *      print depth used by the "print" command.  With an integer 
 *      argument, it sets the current default print depth.   This 
 *      default print depth can be overridden on any particular 
 *      invocation of the "print" command by using the -depth flag,
 *      e.g., print -depth 10 args....  The default print depth is 
 *      initially 1.
 *      See also:  print
 *
 * Syntax:  default-wme-depth [integer]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Sets the default print depth for the agent.
 *
 *----------------------------------------------------------------------
 */

int DefWmeDepthCmd (ClientData clientData, 
		    Tcl_Interp * interp,
		    int argc, char *argv[])
{
  int depth;

  if (argc == 1)
    {
      sprintf(interp->result, "%d", current_agent(default_wme_depth));
      return TCL_OK;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments, should be: default-wme-depth [integer]";
      return TCL_ERROR;
    }

  if (Tcl_GetInt(interp, argv[1], &depth) == TCL_OK)
    {
      current_agent(default_wme_depth) = depth;
    }
  else
    {
      sprintf(interp->result, 
	      "Expected integer for new default print depth: %s", 
	      argv[1]);
      return TCL_OK;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyAgentCmd --
 *
 *	This is the command procedure for the "destroy-agent" command, 
 *	which destroys a Soar agent and its associated Tcl interpreter.
 *
 * Syntax:  destroy-agent agent-name
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Destroys a Soar agent and its Tcl interpreter.  If this 
 *      command is issued in the interpreter being destroyed, then
 *      control switches to the global Soar agent.
 *
 *----------------------------------------------------------------------
 */

int 
DestroyAgentCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  cons * c;
  char * agent_name;

  if (argc > 2)
    {        
      sprintf(interp->result, "Too many arguments");
      return TCL_ERROR;
    }

  if (argc < 2)
    {
      interp->result = "Expected agent name but none given.";
      return TCL_ERROR;
    }

  agent_name = argv[1];

  for(c = all_soar_agents; c != NIL; c = c->rest) 
    {
      if (!strcmp(agent_name, ((agent *)c->first)->name)) 
	{
	  if (soar_agent == (agent *) c->first)
	    { 
	      sprintf(interp->result, 
		      "Attempt to delete current interpreter (%s) ignored.", 
		      argv[1]);
	      return TCL_ERROR;      	      
	    }	
	  destroy_soar_agent((agent *) c->first);
	  sprintf(interp->result, "Destroying agent %s", agent_name);
	  return TCL_OK;
	}
    }

  /* No agent found, the name doesn't match any agents */
  sprintf(interp->result, "Unknown agent name given: %s.", argv[1]);

  return TCL_ERROR;      

}

/*
 *----------------------------------------------------------------------
 *
 * EchoCmd --
 *
 *      This is the command procedure for the "echo" command, which 
 *      prints text to the currently specified output-strings-destination.
 *      IF logging is enabled, the text is also printed to the log
 *      destination.
 *
 * Syntax:  echo strings
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints the given strings according to the currently specified
 *      output-strings-destination
 *
 *----------------------------------------------------------------------
 */

int EchoCmd (ClientData clientData, 
	     Tcl_Interp * interp,
	     int argc, char *argv[])
{
  int i;
  bool newline = TRUE;

  for (i = 1; i < argc; i++)
    {
      if (string_match_up_to("-nonewline", argv[i], 2))
	{
	  newline = FALSE;
	}
      else 
	{
	  Soar_LogAndPrint((agent *) clientData, argv[i]);
	  if ((i + 1) < argc) 
	    {
	      Soar_LogAndPrint((agent *) clientData, " ");
	    }
	}
    }

  if (newline)
    {
      Soar_LogAndPrint((agent *) clientData, "\n");
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ExciseCmd --
 *
 *      This is the command procedure for the "excise" command, 
 *      which removes productions from Soar's memory.
 *
 *      excise -chunks removes all chunks and justifications
 *      from the agent.  excise -task removes all non-default
 *      productions from the agent and performs an init-soar
 *      command.  excise -all removes all productions from the
 *      agent.  The switches may be abbreviated to 2 characters
 *      (e.g, -c for chunks).
 *
 * Syntax:  excise [production-name | -chunks | -task | -all]*
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Removes the indicated productions from the agent memory.
 *
 *----------------------------------------------------------------------
 */

int ExciseCmd (ClientData clientData, 
	       Tcl_Interp * interp,
	       int argc, char *argv[])
{
  int i;
  production * prod;

  if (argc == 1)
    {
      interp->result = "No arguments given.\nUsage: excise [production-name | -chunks | -task | -all]*";
      return TCL_ERROR;
    }

  for (i = 1; i < argc; i++)
    {
      if (string_match_up_to(argv[i], "-chunks", 2))
	{
	  excise_all_productions_of_type (CHUNK_PRODUCTION_TYPE);
	  excise_all_productions_of_type (JUSTIFICATION_PRODUCTION_TYPE);
	}
      else if (string_match_up_to(argv[i], "-task", 2))
	{
	  excise_all_productions_of_type (USER_PRODUCTION_TYPE);
	  excise_all_productions_of_type (CHUNK_PRODUCTION_TYPE);
	  excise_all_productions_of_type (JUSTIFICATION_PRODUCTION_TYPE);
	  reinitialize_soar();  /* for excise -task, also do an init-soar */
	}
      else if (string_match_up_to(argv[i], "-all", 2))
	{
	  excise_all_productions_of_type (DEFAULT_PRODUCTION_TYPE);
	  excise_all_productions_of_type (USER_PRODUCTION_TYPE);
	  excise_all_productions_of_type (CHUNK_PRODUCTION_TYPE);
	  excise_all_productions_of_type (JUSTIFICATION_PRODUCTION_TYPE);
	  reinitialize_soar();  /* for excise -all, also do an init-soar */
	}
      else if ((prod = name_to_production(argv[i])))
	{
	  excise_production(prod, TRUE);
	}
      else
	{
	  sprintf(interp->result, 
		  "Unrecognized argument to excise: %s", argv[i]);
	  return TCL_ERROR;
	}
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ExplainCmd --
 *
 *      This is the command procedure for the "explain" command.
 *
 *      Explain provides some interpretation of backtraces generated 
 *      during chunking.  Explain mode must be ON when the chunk/
 *      justification is CREATED or no explanation will be available.  
 *      When explain mode is on, more memory is used, and building 
 *      chunks/justifications will be slower.
 *
 *      The two most useful commands are 'explain <name>' and
 *      'explain <name> <cond-num>'.
 *      The first command lists all of the conditions for the named
 *      chunk/justification, and the 'ground' which resulted in
 *      inclusion in the chunk/justification.  A 'ground' is a WME 
 *      which was tested in the supergoal.  Just knowing which WME was 
 *      tested may be enough to explain why the chunk/justification 
 *      exists.  If not, the conditions can be listed with an integer 
 *      value.  This value can be used in 'explain <name> <cond-num>' 
 *      to obtain a list of the productions which fired to obtain this 
 *      condition in the chunk/justification (and the crucial WMEs 
 *      tested along the way).  Why use an integer value to specify the 
 *      condition?  To save a big parsing job. 
 *
 * Syntax: explain arg
 *         arg := ''                 list chunks/justifications if 
 *                                   explain is on
 *         arg := <name>             list all conditions & grounds for 
 *                                   a chunk/justification
 *         arg := <name> -trace      give the backtrace for a named
 *                                   chunk/justification
 *         arg := <name> <cond-num>  explain why this condition is in 
 *                                   the chunk/justification
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int ExplainCmd (ClientData clientData, 
		Tcl_Interp * interp,
		int argc, char *argv[])
{
  if (argc == 1)
    {
      explain_list_chunks();
      return TCL_OK;
    }

  {
    int cond_num;
    
    get_lexeme_from_string(argv[1]);

    if (current_agent(lexeme).type==SYM_CONSTANT_LEXEME) 
      {
	if (argc > 2) 
	  {
	    if (string_match("-trace", argv[2]))
	      {
		/* handle the 'explain name -trace' case */
		
		explain_trace_named_chunk(current_agent(lexeme.string));
	      }
	    else if (Tcl_GetInt(interp, argv[2], &cond_num) == TCL_OK)
	      {
		/* handle the 'explain name <cond-num>' case */

		explain_chunk(current_agent(lexeme.string), cond_num);
	      }
	    else
	      {
		sprintf(interp->result,
			"Unexpected argument to %s %s: %s.  Should be -trace or integer.",
			argv[0], argv[1], argv[2]);
		return TCL_ERROR;
	      }      
	  }
	else
	  {
	      /* handle the 'explain name' case */
	    
	    explain_cond_list(current_agent(lexeme.string));
	  }
      }
    else
      {
	sprintf(interp->result,
		"Unexpected argument to %s: %s.  Should be symbolic constant or integer.",
		argv[0], argv[1]);
	return TCL_ERROR;
      }
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FiringCountsCmd --
 *
 *      This is the command procedure for the "firing-counts" command, 
 *      which prints out how many times a production has fired.
 *
 *      This command prints how many times certain productions have 
 *      fired.  With no arguments, it lists all the productions sorted 
 *      according to how many times they have fired.  If an integer 
 *      argument (call it k) is given, only the top k productions are 
 *      listed.  If k=0, only the productions which haven't fired at 
 *      all are listed.  Note that firing counts are not reset by an 
 *      (init-soar); the counts indicate the number of firings since the
 *      productions were loaded or built.
 *
 *      NB:  this is slow, because the sorting takes time O(n*log n)
 *
 *      With one or more production names as arguments, this command 
 *      prints how many times each of those productions fired.
 *
 * Syntax:  firing-counts [[integer] | production-name* ]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints information about production firings.
 *
 *----------------------------------------------------------------------
 */

int FiringCountsCmd (ClientData clientData, 
		     Tcl_Interp * interp,
		     int argc, char *argv[])
{
  int num_requested;

  if (argc > 1)
    {
      if (Tcl_GetInt(interp, argv[1], &num_requested) != TCL_OK)
	{
	  int i;

	  interp->result = "";          /* Clear error message from */
                                        /* failed int parse         */
	  for(i = 1; i < argc; i++)
	    {
	      production * p;

	      p = name_to_production(argv[i]);
	      if (p)
		{
		  print ("%6lu:  %s\n", p->firing_count, argv[i]);
		}
	      else
		{
		  sprintf (interp->result,
			   "No production named %s", 
			   argv[i]);
		  return TCL_ERROR;
		}
	    }
	  return TCL_OK;
	}
    }

  print_production_firings(interp, ((argc == 1) ? 20 : num_requested));

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FormatWatchCmd --
 *
 *      This is the command procedure for the "format-watch" 
 *      command, which defines the format to use when printing 
 *      objects and the Soar goal stack.
 *
 *      Object trace formats control how Soar prints an object--
 *      e.g., a certain operator, problem-space, etc.  (This is 
 *      like trace-attributes in Soar 5.)  Stack trace formats 
 *      control how Soar prints its context stack selections
 *      in 'watch 0' and 'pgs' printouts.  You specify a trace 
 *      format by indicating two things:
 *      - a format string, indicating the printout format to be used
 *      - what things this format string can be applied to
 *
 *      The format string can be any string.
 *      Certain 'escape sequences' can be used within the string; 
 *      for example, '%dc' means print the current decision cycle 
 *      number.
 * 
 *      There are two ways to restrict what objects a format string 
 *      applies to.  The {s|o|*} argument restricts the types 
 *      of objects:  's' indicates that the format only applies to 
 *      states; 'o' means it only applies to operators; and
 *      '*' means it applies to any type of object.  The [object-name]
 *      argument (for object trace formats), if given, means it only 
 *      applies to objects with that ^name.  The [ps-name] argument 
 *      (for stack trace formats) means it only applies within problem 
 *      spaces with that ^name.
 * 
 *      With an -add argument, these commands add new trace formats 
 *      (replacing any existing ones with identical applicability 
 *      conditions).  With a -remove argument, they remove trace 
 *      formats with the given applicability conditions.  With no 
 *      arguments, they print out all current trace formats.
 *
 *      The following escape sequences can be used within trace format 
 *      strings.  The S indicates the sequence is ONLY usable in
 *      stack traces:
 *
 *      %cs                S- print the current state
 *      %co                S- print the current operator 
 *                          * The %cg, %cp, %cs, and %co sequences use 
 *                            the appropriate object trace format.
 *      %dc                S- print the current decision cycle number 
 *      %ec                S- print the current elaboration cycle number
 *                          * The %dc and %ec sequences are NOT
 *                          * meaningful in stack traces produced by 
 *                            the "pgs" command.  In these cases, nothing 
 *                            is printed.
 *      %sd                S- print the current subgoal depth
 *                          * The %sd sequence uses 0 as the top level).
 *      %rsd[pattern]      S- repeat (subgoal depth) times: print the 
 *                            given pattern.
 *      %left[num,pattern]  - print the pattern left justified in a
 *                            field of num spaces.
 *      %right[num,pattern] - print the pattern right justified in a
 *                            field of num spaces.
 *      %id                 - print the identifier of the current object.
 *      %v[foo]             - print the value(s) of attribute ^foo on the 
 *                            current object.  If there is no ^foo on the 
 *                            current object, nothing is printed.
 *      %v[foo.bar.baz]     - same as the above, only follow the given 
 *                            attribute path to get the value(s).
 *      %v[*]               - print all values of all attributes on the 
 *                            current object.
 *      %o[args]            - same as %v, except that if the value is an 
 *                            identifier it is printed using the 
 *                            appropriate object trace format.
 *      %av[args]           - same as %v, except the printed value is 
 *                            preceeded with "^attr " to indicate the 
 *                            attribute name.
 *      %ao[args]           - a combination of the above two.
 *      %ifdef[pattern]     - print the given pattern if and only if all 
 *                            escape sequences inside it are "meaningful" 
 *                            or "well-defined."  For example, 
 *                            "%ifdef[foo has value: %v[foo]]" will print 
 *                            nothing if there is no ^foo on the current 
 *                            object.
 *      %%                  - print a percent sign
 *      %[                  - print a left bracket
 *      %]                  - print a right bracket.
 *
 * Syntax:
 *      format-watch {-object | -stack} 
 *                   [{ { -add    {s|o|*} [name] format}
 *                   |  { -remove {s|o|*} [name]       }
 *                    }
 *                   ]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Sets the indicated format.
 *
 *----------------------------------------------------------------------
 */

int FormatWatchCmd (ClientData clientData, 
		    Tcl_Interp * interp,
		    int argc, char *argv[])
{
  static char * too_few_args  = "Too few arguments.\nUsage: format-watch {-object | -stack} [{{ -add {s|o|*} [name] \"format\" }|{-remove {s|o|*} [name]}}]";
  static char * too_many_args = "Too many arguments.\nUsage: format-watch {-object | -stack} [{{ -add {s|o|*} [name] \"format\" }|{-remove {s|o|*} [name]}}]";

  bool stack_trace;
  int type_restriction;
  Symbol *name_restriction = NIL;
  bool remove;
  int format_arg = 0;  /* Initialized to placate gcc -Wall */

  if (argc == 1)
    {
      interp->result = too_few_args;
      return TCL_ERROR;
    }

  /* --- set stack_trace depending on which option was given --- */  
  if (string_match("-stack", argv[1]))
    {
      stack_trace = TRUE;
    }
  else if (string_match("-object", argv[1]))
    {
      stack_trace = FALSE;
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized option to format-watch : %s",
	      argv[1]);
      return TCL_ERROR;
    }

  /* --- if no further args, print all trace formats of that type --- */
  
  if (argc == 2)
    {
      print_all_trace_formats_tcl (stack_trace);
      return TCL_OK;
    }

  /* --- next argument must be either -add or -remove --- */
  if (string_match("-add", argv[2]))
    {
      remove = FALSE;
    }
  else if (string_match("-remove", argv[2]))
    {
      remove = TRUE;
    }
  else 
    {
      sprintf(interp->result,
	      "Unrecognized option to format-watch %s: %s",
	      argv[1], argv[2]);
      return TCL_ERROR;
    }

  if (argc == 3)
    {
      interp->result = too_few_args;
      return TCL_ERROR;
    }

  /* --- read context item argument: s, o, or '*' --- */
  
  if (string_match("s", argv[3]))
      {
	type_restriction = FOR_STATES_TF;
      }
  else if (string_match("o", argv[3]))
      {
	type_restriction = FOR_OPERATORS_TF;
      }
  else if (string_match("*", argv[3]))
      {
	type_restriction = FOR_ANYTHING_TF;
      }
  else 
      {
	sprintf(interp->result,
		"Unrecognized option to %s %s %s: %s",
		argv[0], argv[1], argv[2], argv[3]);
	return TCL_ERROR;	
      }

  if (argc > 4)
    {
      get_lexeme_from_string(argv[4]);

      /* --- read optional name restriction --- */
      if (current_agent(lexeme).type == SYM_CONSTANT_LEXEME) 
	{
	  name_restriction = make_sym_constant (argv[4]);
	  format_arg = 5;
        }
      else
	{
	  format_arg = 4;
	}

      if (   (remove  && (argc > format_arg))
	  || (!remove && (argc > (format_arg + 1))))
	{
	  interp->result = too_many_args;
	  return TCL_ERROR;
	}
    }

  /* --- finally, execute the command --- */
  if (remove) 
    {
      remove_trace_format (stack_trace, type_restriction, name_restriction);
    } 
  else 
    {
      if (argc == (format_arg + 1))
	{
	  add_trace_format (stack_trace, 
			    type_restriction, 
			    name_restriction,
			    argv[format_arg]);
	}
      else
	{
	  if (name_restriction)
	    {
	      symbol_remove_ref (name_restriction);
	    }
	  
	  sprintf(interp->result,
		  "Missing format string");

	  return TCL_ERROR;
	}
    }

  if (name_restriction) 
    {
      symbol_remove_ref (name_restriction);
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GoCmd --
 *
 *      This is the command procedure for the "go" command which
 *      runs the Soar agent.
 * 
 *       This is the most general command for running Soar.  It 
 *       takes two optional arguments, one specifying how many 
 *       things to run, and one specifying what type of things to 
 *       run.  The following types are available:
 * 
 *       p  - run Soar for n phases.  A phase is either an input 
 *            phase, preference phase, working memory phase, output 
 *            phase, or decision phase.
 *       e  - run Soar for n elaboration cycles.  (For purposes of 
 *            this command, decision phase is counted as an 
 *            elaboration cycle.)
 *       d  - run Soar for n decision cycles.
 *       s  - run Soar until the nth time a state is selected.
 *       o  - run Soar until the nth time an operator is selected.
 *       context-variable - run Soar until the nth time a selection 
 *            is made for that particular context slot, or until the 
 *            context stack pops to above that context.
 * 
 *       Go remembers each argument you give it each time.  If you 
 *       don't give arguments next time, it uses the ones from the 
 *       previous time.
 * 
 *       Examples:
 *         go 5 d   --> run for 5 decision cycles
 *         go e     --> run for another 5 elaboration cycles
 *         go 1 s   --> run until the next state is selected (i.e., 
 *                      until the next time an impasse arises)
 *         go <so>  --> run until the next superoperator is selected 
 *                      (or until the supergoal goes away)
 *         go 3 <o> --> run for 3 operator selections at this level 
 *                      (continuing through any subgoals that arise)
 *
 * Syntax:  go [integer | 'forever'] [type]
 *          type ::= p | e | d | s | o | context-variable
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Runs the Soar agent for the indicated duration
 *
 *----------------------------------------------------------------------
 */

int GoCmd (ClientData clientData, 
	   Tcl_Interp * interp,
	   int argc, char *argv[])
{
  agent * the_agent;
  int parse_result;
  long               go_number;
  enum go_type_enum  go_type;
  Symbol *           go_slot_attr;
  goal_stack_level   go_slot_level;
  
  static bool executing = FALSE;

  if (executing == TRUE) 
    {
      /* 
       * Disallow attempts to recursively enter go-related actions.
       * This would lead to seg faults since the agent code is not
       * re-entrant.  Note that this is a general problem, so this
       * does not solve the problem in a general fashion.  This
       * strategy should probably be used on long running commands,
       * however, unless a more general solution is found.
       *
       * This possibility is easy to generate when "go" is tied to
       * a button on the GUI and the user clicks faster than "go"
       * can run.
       */

      return TCL_OK;
    }
  else
    {
      executing = TRUE;
    }

  the_agent = (agent *) clientData;

  go_number     = the_agent->go_number;
  go_type       = the_agent->go_type;
  go_slot_attr  = the_agent->go_slot_attr;
  go_slot_level = the_agent->go_slot_level;

  parse_result = parse_go_command(interp, argc, argv,
				  &go_number, &go_type,
				  &go_slot_attr, &go_slot_level);
  if (parse_result == TCL_OK) 
    {
      the_agent->go_number     = go_number;
      the_agent->go_type       = go_type;
      the_agent->go_slot_attr  = go_slot_attr;
      the_agent->go_slot_level = go_slot_level;
      
      execute_go_selection(the_agent, go_number, go_type,
			   go_slot_attr, go_slot_level);
    }
  else
    {
      executing = FALSE;
      return TCL_ERROR;
    }

  interp->result = "\n";
  executing = FALSE;
  return parse_result;
}

/*
 *----------------------------------------------------------------------
 *
 * InitSoarCmd --
 *
 *      This is the command procedure for the "init-soar" command, 
 *      (re)initializes the Soar agent.
 * 
 *      This command re-initializes Soar.  It removes all wmes from 
 *      working memory, wiping out the goal stack, and resets all 
 *      statistics (except the counts of, how many times each 
 *      individual production has fired, used by the "firing-counts" 
 *      command).
 *
 * Syntax:  init-soar
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Empties working memory, removes the goal stack, and resets
 *      statistics.
 *
 *----------------------------------------------------------------------
 */

int InitSoarCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  if (argc > 1)
    {
      interp->result = "Too many arguments, should be: init-soar";
      return TCL_ERROR;
    }

  reinitialize_soar();
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * InputPeriodCmd --
 *
 *      This is the command procedure for the "input-period" 
 *      command, which sets/prints the Soar input period.
 *
 *      With no arguments, this command prints the current period
 *      used to control the input rate to the Soar agent.  With an
 *      integer argument, it sets the current input period.  If
 *      the argument is 0, Soar behaves as it did in versions before
 *      6.2.4, accepting input every elaboration cycle.  An input
 *      period of "n" means that input is accepted only every "n"
 *      decision cycles.  The input period is initially set to 0.
 *      Negative integer values are disallowed.
 *
 * Syntax:  input-period [integer]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Sets/prints the input period for the agent.
 *
 *----------------------------------------------------------------------
 */

int InputPeriodCmd (ClientData clientData, 
		    Tcl_Interp * interp,
		    int argc, char *argv[])
{
  int period;

  if (argc == 1)
    {
      sprintf(interp->result, "%d", current_agent(input_period));
      return TCL_OK;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments, should be: input-period [integer]";
      return TCL_ERROR;
    }

  if (Tcl_GetInt(interp, argv[1], &period) == TCL_OK)
    {
      if (period >= 0)
	{
	  current_agent(input_period) = period;
	}
      else
	{
	  sprintf(interp->result, 
		  "Integer for new input period must be >= 0, not %s", 
		  argv[1]);
	  return TCL_ERROR;
	}
    }
  else
    {
      sprintf(interp->result, 
	      "Expected integer for new input period: %s", 
	      argv[1]);
      return TCL_ERROR;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * InternalSymbolsCmd --
 *
 *      This is the command procedure for the "internal-symbols"
 *      command which prints information about the Soar agent symbol
 *      table.
 *
 * Syntax:  internal-symbols
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints the current Soar agent symbol table
 *
 *----------------------------------------------------------------------
 */
int InternalSymbolsCmd (ClientData clientData, 
			Tcl_Interp * interp,
			int argc, char *argv[])
{
  if (argc > 1)
    {
      interp->result = "Too many arguments, should be: internal-symbols";
      return TCL_ERROR;
    }

  print_internal_symbols();

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * LearnCmd --
 *
 *      This is the command procedure for the "learn" command, which 
 *
 *      With no arguments, this command prints out the current learning 
 *      status.  Any of the following arguments may be given:
 *
 *        on         - turns learning on
 *        off        - turns all learning off
 *        specify    - learns in chunky problem spaces  * AGR MVL1 *
 *        all-goals  - when learning is on, this allows learning at all 
 *                     goal stack levels (in contrast to bottom-up 
 *                     learning)
 *        bottom-up  - when learning is on, this allows learning at only 
 *                     the lowest goal stack level; i.e., a chunk is 
 *                     learned at a given level only if no chunk has yet 
 *                     been learned at a lower level.
 *        noprint    - equivalent to (watch -chunk-names off -chunks off)
 *        print      - equivalent to (watch -chunk-names  on -chunks off)
 *        full-print - equivalent to (watch -chunk-names  on 
 *                                          -chunks  on
 *                                          -justification-names on
 *                                          -justifications  on)
 *        notrace    - equivalent to (watch -firings chunk off 
 *                                          -backtracing off)
 *        trace      - equivalent to (watch -firings chunk on 
 *                                          -backtracing off)
 *        full-trace - equivalent to (watch -firings chunk on 
 *                                          -backtracing on
 *                                          -chunk-names on 
 *                                          -chunks on
 *                                          -justification-names on 
 *                                          -justifications on)
 *
 * See also: chunk-free-problem-spaces, watch
 *
 * Syntax:  learn arg*
 *            arg  ::=  -on | -specify | -off 
 *            arg  ::=  -all-goals | -bottom-up
 *            arg  ::=  -noprint | -print | -full-print
 *            arg  ::=  -notrace | -trace | -full-trace
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Runs the agent for the indicated number of decision cycles.
 *
 *----------------------------------------------------------------------
 */

int LearnCmd (ClientData clientData, 
	      Tcl_Interp * interp,
	      int argc, char *argv[])
{
  if (argc == 1)
    {
      print_current_learn_settings();
      return TCL_OK;
    }

  {int i;

   for (i = 1; i < argc; i++)
     {
       if (string_match("-on", argv[i]))
	 {
	   set_sysparam (LEARNING_ON_SYSPARAM, TRUE); 
	   set_sysparam (LEARNING_SPECIFY_SYSPARAM, FALSE);  /* AGR MVL1 */
	   set_sysparam (TRACE_CHUNK_NAMES_SYSPARAM, TRUE);
	   set_sysparam (TRACE_CHUNKS_SYSPARAM, FALSE);
	   set_sysparam (TRACE_FIRINGS_OF_CHUNKS_SYSPARAM, TRUE);
	   set_sysparam (TRACE_BACKTRACING_SYSPARAM, FALSE);
	 }
       else if (string_match_up_to("-specify", argv[i], 3))
	 {
/* AGR MVL1 begin */
	set_sysparam (LEARNING_ON_SYSPARAM, TRUE); 
	set_sysparam (LEARNING_SPECIFY_SYSPARAM, TRUE);
	set_sysparam (TRACE_CHUNK_NAMES_SYSPARAM, TRUE);
	set_sysparam (TRACE_CHUNKS_SYSPARAM, FALSE);
	set_sysparam (TRACE_FIRINGS_OF_CHUNKS_SYSPARAM, TRUE);
	set_sysparam (TRACE_BACKTRACING_SYSPARAM, FALSE);
/* AGR MVL1 end */
	 }
       else if (string_match_up_to("-off", argv[i], 3))
	 {
	   set_sysparam (LEARNING_ON_SYSPARAM, FALSE);
	   set_sysparam (LEARNING_SPECIFY_SYSPARAM, FALSE);  /* AGR MVL1 */
	   set_sysparam (TRACE_CHUNK_NAMES_SYSPARAM, FALSE);
	   set_sysparam (TRACE_CHUNKS_SYSPARAM, FALSE);
	   set_sysparam (TRACE_FIRINGS_OF_CHUNKS_SYSPARAM, FALSE);
	   set_sysparam (TRACE_BACKTRACING_SYSPARAM, FALSE);
	 }
       else if (string_match_up_to("-all-goals", argv[i], 2)) 
	 {
	   set_sysparam (LEARNING_ALL_GOALS_SYSPARAM, TRUE);
	 }
       else if (string_match_up_to("-bottom-up", argv[i], 2))
	 {
	   set_sysparam (LEARNING_ALL_GOALS_SYSPARAM, FALSE);
	 }
       else if (string_match_up_to("-noprint", argv[i], 4))
	 {
	   set_sysparam (TRACE_CHUNK_NAMES_SYSPARAM, FALSE);
	   set_sysparam (TRACE_CHUNKS_SYSPARAM, FALSE);
	 }
       else if (string_match_up_to("-print", argv[i], 2))
	 {
	   set_sysparam (TRACE_CHUNK_NAMES_SYSPARAM, TRUE);
	   set_sysparam (TRACE_CHUNKS_SYSPARAM, FALSE);
	 }
       else if (string_match_up_to("-full-print", argv[i], 7))
	 {
	   set_sysparam (TRACE_CHUNK_NAMES_SYSPARAM, TRUE);
	   set_sysparam (TRACE_CHUNKS_SYSPARAM, TRUE);
	   set_sysparam (TRACE_JUSTIFICATION_NAMES_SYSPARAM, TRUE);
	   set_sysparam (TRACE_JUSTIFICATIONS_SYSPARAM, TRUE);
	 }
       else if (string_match_up_to("-notrace", argv[i], 4))
	 {
	   set_sysparam (TRACE_FIRINGS_OF_CHUNKS_SYSPARAM, FALSE);
	   set_sysparam (TRACE_BACKTRACING_SYSPARAM, FALSE);
	 }
       else if (string_match_up_to("-trace", argv[i], 2))
	 {
	   set_sysparam (TRACE_FIRINGS_OF_CHUNKS_SYSPARAM, TRUE);
	   set_sysparam (TRACE_BACKTRACING_SYSPARAM, FALSE);
	 }
       else if (string_match_up_to("-full-trace", argv[i], 7))
	 {
	   set_sysparam (TRACE_FIRINGS_OF_CHUNKS_SYSPARAM, TRUE);
	   set_sysparam (TRACE_BACKTRACING_SYSPARAM, TRUE);
	   set_sysparam (TRACE_CHUNK_NAMES_SYSPARAM, TRUE);
	   set_sysparam (TRACE_CHUNKS_SYSPARAM, TRUE);
	   set_sysparam (TRACE_JUSTIFICATION_NAMES_SYSPARAM, TRUE);
	   set_sysparam (TRACE_JUSTIFICATIONS_SYSPARAM, TRUE);
	 }
       else
	 {
	   sprintf(interp->result,
		   "Unrecognized argument to learn command: %s",
		   argv[i]);
	   return TCL_ERROR;
	 }
     }
 }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ListAgentsCmd --
 *
 *	This is the command procedure for the "list-agents" command, 
 *	which lists all Soar agent names.
 *
 * Syntax: list-agents
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Returns a list of Soar agent names in interp->result.
 *
 *----------------------------------------------------------------------
 */

int 
ListAgentsCmd (ClientData clientData, 
	       Tcl_Interp * interp,
	       int argc, char *argv[])
{
  cons * c;

  if (argc > 1)
    {        
      sprintf(interp->result, "Too many arguments.  Usage: list-agents");
      return TCL_ERROR;
    }

  for(c = all_soar_agents; c != NIL; c = c->rest) 
    {
      Tcl_AppendElement(interp, ((agent *)c->first)->name);
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ListProdCmd --
 *
 *      This is the command procedure for the "list-productions" 
 *      command, which lists productions of various kinds.  
 *
 *      list-chunks is shorthand for list-productions chunk and
 *      list-justifications is shorthand for list-productions 
 *      justication.  If no production types are given, all 
 *      productions except justifications are printed.
 * 
 *      The optional [-internal] argument tells Soar to print 
 *      productions in their internal reordered (rete net) form.
 * 
 * Syntax: list-productions [prod-type*]
 *         prod-type ::= -default |
 *                       -user    | 
 *                       -chunk   | 
 *                       -justification | 
 *                       -internal
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints the indicated types of productions.
 *
 *----------------------------------------------------------------------
 */

int ListProdCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  bool prod_type_present, internal;
  bool types[NUM_PRODUCTION_TYPES]; /* tells which types of prod's to list */
  int i;

  internal = FALSE;
  for (i=0; i<NUM_PRODUCTION_TYPES; i++) 
    types[i]=FALSE; /* init to FALSE */
  
  /* --- read prod-type* --- */
  prod_type_present = FALSE;
  for (i = 1; i < argc; i++)
    {
      if (string_match("-default", argv[i]))
	{
	  prod_type_present = TRUE;
	  types[DEFAULT_PRODUCTION_TYPE] = TRUE;
	}
      else if (string_match("-user", argv[i]))
	{
	  prod_type_present = TRUE;
	  types[USER_PRODUCTION_TYPE] = TRUE;
	}
      else if (string_match("-chunk", argv[i]))
	{
	  prod_type_present = TRUE;
	  types[CHUNK_PRODUCTION_TYPE] = TRUE;
	}
      else if (string_match("-justification", argv[i]))
	{
	  prod_type_present = TRUE;
	  types[JUSTIFICATION_PRODUCTION_TYPE] = TRUE;
	}
      else if (string_match("-internal", argv[i]))
	{
	  internal = TRUE;
	}
      else
	{
	  sprintf(interp->result,
		  "Unrecognized option to list-productions command: %s", 
		  argv[i]);
	  return TCL_ERROR;
	}
    }

  if (!prod_type_present) 
    {    /* --- no prod-type* was present --- */
    types[DEFAULT_PRODUCTION_TYPE] = TRUE;
    types[USER_PRODUCTION_TYPE] = TRUE;
    types[CHUNK_PRODUCTION_TYPE] = TRUE;
    }

  /* --- all set, now print all the productions --- */
  {
    production * prod;

    /* AGR 641 begin */
    /* AGR 641  We want to change it so chunks are printed out starting 
       with the lowest numbered chunks.  That means we need to go to the 
       end of the list and go back.  94.11.02 */

    for (i=0; i<NUM_PRODUCTION_TYPES; i++) {
      if (types[i]) {
	/* first go to end of list to reverse the order  GAP */
	for (prod=current_agent(all_productions_of_type)[i]; 
	     prod != NIL && prod->next!=NIL; 
	     prod=prod->next)
	  /* intentionally null */ ;

	while (prod != NIL){
	  print_production (prod, internal);
	  print ("\n");
	  prod=prod->prev;
	}
      }  /* if type is to be outputted */
    }  /* for all types */
    /* AGR 641 end */
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * LogCmd --
 *
 *      This is the command procedure for the "log" command which 
 *      records session information to a log file.
 *
 *      This command may be used to open and close log files.  With
 *      the -add argument, specific strings can also be added to 
 *      the log file.
 *
 * Syntax:  log <action>
 *          <action> ::= [-new | -existing] pathname 
 *          <action> ::= -add string
 *          <action> ::= -query
 *          <action> ::= -off
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Opens and/or closes log files.
 *
 *----------------------------------------------------------------------
 */

int LogCmd (ClientData clientData, 
	    Tcl_Interp * interp,
	    int argc, char *argv[])
{
  char * too_few = "Too few arguments, should be: log [-new | -existing] pathname | log -add string | log -query | log -off";
  char * too_many = "Too many arguments, should be: log [-new | -existing] pathname | log -add string | log -query | log -off";

  if (argc < 2)
    {
      interp->result = "The log file is ";
      if (soar_exists_callback((soar_callback_agent) clientData,
			       LOG_CALLBACK))
	{
	  Tcl_AppendResult(interp, 
			   "open.  Use log -off to close the file.", 
			   (char *) NULL); 
	}
      else
	{
	  Tcl_AppendResult(interp, "closed.", (char *) NULL); 
	}
      Tcl_AppendResult(interp, "", (char *) NULL);

      return TCL_OK;      
    }

  if (argc > 3)
    {
      interp->result = too_many;
      return TCL_ERROR;      
    }

  if (string_match_up_to("-add", argv[1], 2))
    {
      if (argc == 3)
	{
	  Soar_Log((agent *) clientData, argv[2]);
	}
      else if (argc < 3)
	{
	  interp->result = too_few;
	  return TCL_ERROR;      
	}
      else 
	{
	  interp->result = too_many;
	  return TCL_ERROR;      
	}
    }
  else if (string_match_up_to("-query", argv[1], 2))
    {
      if (argc == 2)
	{
	  if (soar_exists_callback((soar_callback_agent) clientData,
				   LOG_CALLBACK))
	    {
	      interp->result = "open";
	    }
	  else
	    {
	      interp->result = "closed";
	    }
	}
      else if (argc < 2)
	{
	  interp->result = too_few;
	  return TCL_ERROR;      
	}
      else 
	{
	  interp->result = too_many;
	  return TCL_ERROR;      
	}
    }
  else if (string_match_up_to("-off", argv[1], 2))
    {
      if (argc == 2)
	{
	  if (soar_exists_callback((soar_callback_agent) clientData,
				   LOG_CALLBACK))
	    {
	      Soar_Log((agent *) clientData, "\n**** log closed ****\n");
	      soar_pop_callback((soar_callback_agent) clientData,
				LOG_CALLBACK);
	    }
	  else
	    {
	      interp->result = "Attempt to close non-existant log file";
	      return TCL_ERROR;
	    }
	}
      else if (argc > 2)
	{
	  interp->result = too_many;
	  return TCL_ERROR;      
	}
    }
  else
    { /* Either we have a file open/append request or there is an error */
      
      char * filename;
      char * mode;

      if (argc == 2)
	{
	  filename = argv[1];
	  mode = "w";
	} 
      else if (string_match_up_to("-new", argv[1], 2))
	{
	  filename = argv[2];
	  mode = "w";
	}
      else if (string_match_up_to("-new", argv[2], 2))
	{
	  filename = argv[1];
	  mode = "w";
	}
      else if (string_match_up_to("-existing", argv[1], 2))
	{
	  filename = argv[2];
	  mode = "a";
	}
      else if (string_match_up_to("-existing", argv[2], 2))
	{
	  filename = argv[1];
	  mode = "a";
	}
      else
	{
	  sprintf(interp->result,
		  "log: unrecognized arguments: %s %s",
		  argv[1], argv[2]);
	  return TCL_ERROR;
	}

      {
	Tcl_DString buffer;
	char * fullname;
	FILE * f;

	fullname = Tcl_TildeSubst(interp, filename, &buffer);
	if (fullname == NULL)
	  {
	    return TCL_ERROR;
	  }

	if (!(f = fopen(fullname, mode)))
	  {
	    Tcl_DStringFree(&buffer);
	    sprintf (interp->result,
		     "log: Error: unable to open '%s'",
		     filename);
	    return TCL_ERROR;
	  }
	else
	  {
	    /* Install callback */
	    soar_push_callback((soar_callback_agent) clientData, 
			       LOG_CALLBACK,
			       (soar_callback_fn) Soar_PrintToFile,
			       (soar_callback_data) f,
			       (soar_callback_free_fn) Soar_FClose);
	    Soar_Log((agent *) clientData, "**** log opened ****\n");
	  }
      }
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * MatchesCmd --
 *
 *      This is the command procedure for the "matches" command, 
 *      which prints partial match information for a selected
 *      production.  It also prints information about the current
 *      match set.
 *
 *      The match set is a list of productions that are about to 
 *      fire or retract in the next preference phase.  With no 
 *      arguments, this command prints out the production names for 
 *      both Assertions and Retractions.  The first optional argument
 *      specifies listing of either Assertions or Retractions.
 *
 *      The last optional argument specifies the level of detail wanted:  
 *      -counts (the default for single productions) prints out just 
 *      the partial match counts; -names (the default for match sets) 
 *      prints out just the production names; -timetags also prints
 *      the timetags of wmes matched; and -wmes prints the wmes rather
 *       than just their timetags.
 *
 * Syntax:  matches production-name [ -count | -timetags | -wmes]
 * Syntax:  matches production-name [ 0 | 1 | 2 ]
 * Syntax:  ms [-assertions | -retractions] [-names| -timetags | -wmes]
 * Syntax:  ms [-assertions | -retractions] [ 0 | 1 | 2 ]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints production match information.
 *
 *----------------------------------------------------------------------
 */

int MatchesCmd (ClientData clientData, 
		Tcl_Interp * interp,
		int argc, char *argv[])
{
  production * prod = NULL;
  production * prod_named;
  wme_trace_type wtt = NONE_WME_TRACE;
  ms_trace_type  mst = MS_ASSERT_RETRACT;
  int curr_arg = 1;

  while (curr_arg < argc)
    {
      if (string_match_up_to("-assertions", argv[curr_arg], 2))
	{
	  mst = MS_ASSERT;
	}
      else if (string_match_up_to("-retractions", argv[curr_arg], 2))
	{
	  mst = MS_RETRACT;
	}
      else if (   string_match_up_to("-names", argv[curr_arg], 2)
	       || string_match_up_to("-count", argv[curr_arg], 2)
	       || string_match("0", argv[curr_arg]))
	{
	  wtt = NONE_WME_TRACE;
	}
      else if (   string_match_up_to("-timetags", argv[curr_arg], 2)
	       || string_match("1", argv[curr_arg]))
	{
	  wtt = TIMETAG_WME_TRACE;
	}
      else if (   string_match_up_to("-wmes", argv[curr_arg], 2)
	       || string_match("2", argv[curr_arg]))
	{
	  wtt = FULL_WME_TRACE;
	}
      else if ((prod_named = name_to_production(argv[curr_arg])))
	{
	  prod = prod_named;
	}
      else
	{
	  sprintf(interp->result,
		  "Unrecognized option to matches command: %s", 
		  argv[curr_arg]);
	  return TCL_ERROR;
	}

      curr_arg++;
    }

  if (prod != NULL) 
    {
      print_partial_match_information (prod->p_node, wtt);
    }
  else 
    {
      print_match_set (wtt, mst);
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * MaxChunksCmd --
 *
 *      This is the command procedure for the "max-chunks"  command, 
 *      which sets/prints the maximum number of chunks allowed.
 *
 *      With no arguments, this command prints the current value 
 *      of the system variable "max-chunks".  With an integer argument, 
 *      it sets the current value.   This variable controls the maximum 
 *      number of chunks allowed in a single decision cycle.  After this 
 *      many chunks have been executed, Soar proceeds to decision phase 
 *      even if quiescence hasn't really been reached yet.  (Max-chunks 
 *      is initially 50.)
 *
 * Syntax:  max-chunks [integer]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Sets the maximum chunks allowed.
 *
 *----------------------------------------------------------------------
 */

int MaxChunksCmd (ClientData clientData, 
		  Tcl_Interp * interp,
		  int argc, char *argv[])
{
  int num;

  if (argc == 1)
    {
      sprintf(interp->result, "%ld", 
	      current_agent(sysparams)[MAX_CHUNKS_SYSPARAM]);
      return TCL_OK;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments, should be: max-chunks [integer]";
      return TCL_ERROR;
    }

  if (Tcl_GetInt(interp, argv[1], &num) == TCL_OK)
    {
      set_sysparam (MAX_CHUNKS_SYSPARAM, num);
    }
  else
    {
      sprintf(interp->result, 
	      "Expected integer for new maximum chunks count: %s", 
	      argv[1]);
      return TCL_OK;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * MaxElaborationsCmd --
 *
 *      This is the command procedure for the "max-elaborations" 
 *      command, which sets/prints the maximum elaboration cycles
 *      allowed.
 *
 *      With no arguments, this command prints the current value of 
 *      the system variable "max-elaborations".  With an integer 
 *      argument, it sets the current value.   This variable controls 
 *      the maximum number of elaboration cycles allowed in a single 
 *      decision cycle.  After this many elabloration cycles have been 
 *      executed, Soar proceeds to decision phase even if quiescence
 *      hasn't really been reached yet.  (Max-elaborations is initially 
 *      100.)
 *
 * Syntax:  max-elaborations [integer]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Sets the maximum elaborations allowed.
 *
 *----------------------------------------------------------------------
 */

int MaxElaborationsCmd (ClientData clientData, 
			Tcl_Interp * interp,
			int argc, char *argv[])
{
  int num;

  if (argc == 1)
    {
      sprintf(interp->result, "%ld", 
	      current_agent(sysparams)[MAX_ELABORATIONS_SYSPARAM]);
      return TCL_OK;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments, should be: max-elaborations [integer]";
      return TCL_ERROR;
    }

  if (Tcl_GetInt(interp, argv[1], &num) == TCL_OK)
    {
      set_sysparam (MAX_ELABORATIONS_SYSPARAM, num);
    }
  else
    {
      sprintf(interp->result, 
	      "Expected integer for new maximum elaborations count: %s", 
	      argv[1]);
      return TCL_OK;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * MemoriesCmd --
 *
 *      This is the command procedure for the "memories" command, 
 *      which prints information about memory use, in tokens, of 
 *      partial matches of productions.  If a production-name is 
 *      given, memory use for that production is printed.  If no 
 *      production name is given, memories will list 'count' 
 *      productions of the specified type (or all types, if no 
 *      type is specified).  If 'count' is omitted, memory use of
 *      all productions is printed.
 *
 * Syntax:  memories [arg*]
 *          arg  ::=  -chunk | -user | -default | -justification 
 *                           | production-name | count
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints information about memory usage of partial matches.
 *
 *----------------------------------------------------------------------
 */

int MemoriesCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  int i;
  int num;
  int num_items = 20;
  int mems_to_print[NUM_PRODUCTION_TYPES];
  bool set_mems = FALSE;
  production * prod;

  for (i = 0; i < NUM_PRODUCTION_TYPES; i++)
    mems_to_print[i] = FALSE;

  for (i = 1; i < argc; i++)
    {
      if (string_match_up_to(argv[i], "-chunks", 2))
	{
	  mems_to_print[CHUNK_PRODUCTION_TYPE] = TRUE;
	  set_mems = TRUE;
	}
      else if (string_match_up_to(argv[i], "-user", 2))
	{
	  mems_to_print[USER_PRODUCTION_TYPE] = TRUE;
	  set_mems = TRUE;
	}
      else if (string_match_up_to(argv[i], "-default", 2))
	{
	  mems_to_print[DEFAULT_PRODUCTION_TYPE] = TRUE;
	  set_mems = TRUE;
	}
      else if (string_match_up_to(argv[i], "-justification", 2))
	{
	  mems_to_print[JUSTIFICATION_PRODUCTION_TYPE] = TRUE;
	  set_mems = TRUE;
	}
      else if (Tcl_GetInt(interp, argv[i], &num) == TCL_OK)
	{
	  if (num <= 0)
	    {
	      sprintf(interp->result, 
		      "Count argument to memories must be a positive integer, not: %s", 
		      argv[i]);
	      return TCL_ERROR;
	    }
	  else
	    {
	      num_items = num;
	    }
	}
      else if ((prod = name_to_production(argv[i])))
	{
	  print("\n Memory use for %s: %ld\n\n", argv[i], 
		count_rete_tokens_for_production(prod));
	  set_mems = TRUE;
  	}
      else
	{
	  sprintf(interp->result, 
		  "Unrecognized argument to memories: %s", argv[i]);
	  return TCL_ERROR;
	}
    }

  if (!set_mems) {
    mems_to_print[JUSTIFICATION_PRODUCTION_TYPE] = TRUE;
    mems_to_print[CHUNK_PRODUCTION_TYPE] = TRUE;
    mems_to_print[USER_PRODUCTION_TYPE] = TRUE;
    mems_to_print[DEFAULT_PRODUCTION_TYPE] = TRUE;
  }

  print_memories(num_items, mems_to_print);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * MonitorCmd --
 *
 *      This is the command procedure for the "monitor" command, 
 *      which manages the attachment of scripts to Soar events.
 *
 * Syntax:  monitor <add-specification> | 
 *                  <removal-specification> |
 *                  <list-specification> |
 *                  -test |
 *                  -clear
 *
 *          <add-specification>     := -add soar-event script [id]
 *          <removal-specification> := -delete soar-event [id]
 *          <list-specification>    := -list [soar-event]
 *
 * Results:
 *      Returns a standard Tcl completion code. 
 *
 * Side effects:
 *      Adds and/or removes indicated monitor from the system.
 *      This command may also print information about defined
 *      monitors.  Returns the name of the new monitor attachments
 *      if one is created.
 *
 *----------------------------------------------------------------------
 */

static int monitor_counter = 1;

int MonitorCmd (ClientData clientData, 
		Tcl_Interp * interp,
		int argc, char *argv[])
{
  static char * too_few_args_string = "Too few arguments, should be: monitor [-add event script [id]] | [-delete event [id]] | [-list [event] | clear]";
  static char * too_many_args_string = "Too many arguments, should be: monitor [-add event script [id]] | [-delete event [id]] | [-list [event] | -clear]";
  char * monitor_id;
  char   buff[10];          /* What size is good here? */

  if (argc == 1)
    {
      interp->result = too_few_args_string;
      return TCL_ERROR;
    }

  if (string_match_up_to(argv[1], "-add", 2))
    {
      if (argc < 4)
	{
	  interp->result = too_few_args_string;
	  return TCL_ERROR;
	}

      if (argc > 5)
	{
	  interp->result = too_many_args_string;
	  return TCL_ERROR;
	}

      if (argc == 5)
	{
	  monitor_id = argv[4];
	}
      else
	{
	  sprintf(buff, "m%d", monitor_counter++);
	  monitor_id = buff;
	}

      {
	SOAR_CALLBACK_TYPE ct;
	
	ct = soar_callback_name_to_enum(argv[2], TRUE);
	if (ct)
	  {
	    soar_add_callback((agent *) clientData, 
			      ct, 
			      soar_callback_to_tcl, 
			      (soar_callback_data) savestring(argv[3]), 
			      soar_callback_data_free_string,
			      (soar_callback_id) monitor_id);
	    interp->result = monitor_id;
	    return TCL_OK;
	  }
	else
	  {
	    sprintf(interp->result,
		    "Attempt to add unrecognized callback event: %s",
		    argv[2]);
	    return TCL_ERROR;
	  }
      }
    }
  else if (string_match_up_to(argv[1], "-delete", 2))
    {
      switch (argc) {
      case 2:
	interp->result = too_few_args_string;
	return TCL_ERROR;
      case 3:	  /* Delete all callbacks of the given type */
	  {
	    SOAR_CALLBACK_TYPE ct;

	    ct = soar_callback_name_to_enum(argv[2], TRUE);
	    if (ct)
	      {
		soar_remove_all_callbacks_for_event((agent *) clientData, ct);
	      }
	    else
	      {
		sprintf(interp->result,
			"Attempt to delete unrecognized callback event: %s",
			argv[2]);
		return TCL_ERROR;
	      }
	  }
	  break;
      case 4:	  /* Delete single callback for given event */
	  {
	    SOAR_CALLBACK_TYPE ct;

	    ct = soar_callback_name_to_enum(argv[2], TRUE);
	    if (ct)
	      {
		soar_remove_callback((agent *) clientData, ct, argv[3]);
	      }
	    else
	      {
		sprintf(interp->result,
			"Attempt to delete unrecognized callback event: %s",
			argv[2]);
		return TCL_ERROR;
	      }
	  }
	  break;
      default:
	interp->result = too_many_args_string;
	return TCL_ERROR;
      }
    }
  else if (string_match_up_to(argv[1], "-list", 2))
    {
      switch (argc) {
      case 2:	  
	soar_list_all_callbacks((agent *) clientData, TRUE);
	break;
      case 3:	  
	{
	  SOAR_CALLBACK_TYPE ct;
	  
	  ct = soar_callback_name_to_enum(argv[2], TRUE);
	  if (ct)
	    {
	      soar_push_callback((soar_callback_agent) clientData,
				 PRINT_CALLBACK,
				 (soar_callback_fn) Soar_AppendResult, 
				 (soar_callback_data) NULL,
				 (soar_callback_free_fn) NULL);
	      soar_list_all_callbacks_for_event((agent *) clientData, ct);
	      soar_pop_callback((soar_callback_agent) clientData, 
				PRINT_CALLBACK);
	    }
	  else
	    {
	      sprintf(interp->result,
		      "Attempt to list unrecognized callback event: %s",
		      argv[2]);
	      return TCL_ERROR;
	    }
	}
	break;
      default:
	interp->result = too_many_args_string;
	return TCL_ERROR;
      }

    }
  else if (string_match_up_to(argv[1], "-test", 2))
    {
      soar_test_all_monitorable_callbacks((agent *)clientData);
    }
  else if (string_match_up_to(argv[1], "-clear", 2))
    {
      /* Delete all callbacks of all types */
      soar_remove_all_monitorable_callbacks((agent *) clientData);
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized option to monitor command: %s",
	      argv[1]);
      return TCL_ERROR;

    }
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * MultiAttrCmd --
 *
 *      This is the command procedure for the "multi-attribute" command, 
 *      which enables a symbol to have multiple attribute values.
 * 
 *      If given, the value must be a positive integer.  If no value is
 *      given, 100 is used as the default.
 *
 * Syntax: multi-attribute symbol [value]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Defines the symbol to be multi-attributed.
 *
 *----------------------------------------------------------------------
 */

int MultiAttrCmd (ClientData clientData, 
		  Tcl_Interp * interp,
		  int argc, char *argv[])
{
  int num;

  if (argc == 1)
    {
      interp->result = "Too few arguments, should be: multi-attribute symbol [value]";
      return TCL_ERROR;
    }

  if (argc > 3)
    {
      interp->result = "Too many arguments, should be: multi-attribute symbol [value]";
      return TCL_ERROR;
    }

  get_lexeme_from_string(argv[1]);

  if (current_agent(lexeme).type != SYM_CONSTANT_LEXEME)
    {
      sprintf(interp->result,
	      "Expected symbolic constant for symbol but got: %s",
	      argv[1]);      
      return TCL_ERROR;
    }

  if (argc == 3)
    {
      if (Tcl_GetInt(interp, argv[2], &num) != TCL_OK)
	{
	  sprintf(interp->result,
		  "Non-integer given for attribute count: %s",
		  argv[2]);      
	  return TCL_ERROR;
	}
      else
	{
	  if (num > 1)
	    {
	      add_multi_attribute_or_change_value(argv[1], num);
	    } 
	  else 
	    {
	      sprintf(interp->result,
		      "Integer must be greater than 1 but was %s",
		      argv[2]);      
	      return TCL_ERROR;
	    }
	}
    }
  else
    {
      add_multi_attribute_or_change_value(argv[1], 100);
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * OSupportModeCmd --
 *
 *	This is the command procedure for the "o-support-mode" command, 
 *	which controls the way o-support calculations are done (for the
 *      current agent).  It takes a single numeric argument: 
 *          0  --> do o-support calculations the normal (Soar 6) way.
 *          1  --> do o-support calculations the normal (Soar 6) way, but
 *                 print a warning message whenever a preference is created
 *                 that would get different support under Doug's proposal.
 *          2  --> do o-support calculations according to Doug Pearson's
 *                 proposal.  (See osupport.c for details.)
 *
 * Syntax:  o-support-mode {0 | 1 | 2}
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Sets current_agent(o_support_calculation_type).
 *
 *----------------------------------------------------------------------
 */

int OSupportModeCmd (ClientData clientData, 
		     Tcl_Interp * interp,
		     int argc, char *argv[])
{
  if (argc > 2)
    {
      interp->result = "Too many arguments.\nUsage: o-support-mode 0|1|2";
      return TCL_ERROR;
    }

  if (argc == 2) {
    if (! strcmp(argv[1], "0")) {
      current_agent(o_support_calculation_type) = 0;
    } else if (! strcmp(argv[1], "1")) {
      current_agent(o_support_calculation_type) = 1;
    } else if (! strcmp(argv[1], "2")) {
      current_agent(o_support_calculation_type) = 2;
    } else {
      sprintf(interp->result,
	      "Unrecognized argument to %s: %s.  Integer 0, 1, or 2 expected.",
	      argv[0], argv[1]);
      return TCL_ERROR;
    }
  }

  sprintf(interp->result, "%d", current_agent(o_support_calculation_type));
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * OutputStringsDestCmd --
 *
 *      This is the command procedure for the "output-strings-destination"
 *      command which redirects strings printed by Soar_Print to the
 *      selected destination.
 *
 *      If output-strings-destination is set to -append-to-result and 
 *      the C code performs an assignment to interp->result then
 *      the intermediate results will be lost (memory leak?).
 *
 * Syntax:  output-strings-destination [-push [ [-text-widget widget-name 
 *                                                           [interp-name]]
 *                                   | [-file-id open-file-id]
 *                                   | -discard 
 *                                   | -append-to-result 
 *                                   ]
 *                            | -pop ]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Changes the destination of Soar_Print commands.
 *
 *----------------------------------------------------------------------
 */
int OutputStringsDestCmd (ClientData clientData, 
			  Tcl_Interp * interp,
			  int argc, char *argv[])
{
  static char * too_few_args = "Too few arguments, should be: output-strings-destination [ -push [[-text-widget widget-name [interp-name]] | [-file file-name] | -discard | -append-to-result] | -pop]";

  if (argc == 1)
    {
      interp->result = too_few_args;
      return TCL_ERROR;
    }

  if (string_match("-push", argv[1]))
    {
      if (string_match("-text-widget", argv[2]))
#ifdef USE_TK
	{
	  if (argc == 3)
	    {
	      interp->result = too_few_args;
	      return TCL_ERROR;
	    }
	  else
	    {
	      /* We assume that we'll be printing to the same interp. */
	      Tcl_Interp * print_interp = interp;
	      Soar_TextWidgetPrintData * data;

	      if (argc > 5)
		{
		  /* Too many arguments */
		  interp->result = "Too many arguments";
		  return TCL_ERROR;
		}
	      if (argc == 5)
		{
		  /* Prints are to go to a specific interp */
		  RegisteredInterp * riPtr = Soar_GetRegisteredInterp(argv[4]);
		  if (riPtr == NULL) 
		    {
		      sprintf(interp->result,
			      "Cannot redirect prints to non-existant interp: %s\n",
			      argv[4]);
		      return TCL_ERROR;
		    }
		  print_interp = riPtr->interp;
		}

	      data = Soar_MakeTextWidgetPrintData (print_interp, argv[3]);
	      soar_push_callback((soar_callback_agent) clientData, 
				 PRINT_CALLBACK,
				 (soar_callback_fn) Soar_PrintToTextWidget,
				 (soar_callback_data) data,
				 (soar_callback_free_fn) Soar_FreeTextWidgetPrintData);
     	    }
	}
#else
	{
	  interp->result = "output-strings-destination: the -text-widget option is only available when using Tk";
	  return TCL_ERROR;
	}
#endif /* USE_TK */
      else if (string_match("-file-id", argv[2]))
	{
	  if (argc == 3)
	    {
	      interp->result = too_few_args;
	      return TCL_ERROR;
	    }
	  else 
	    {
	      FILE * f;

	      if (Tcl_GetOpenFile(interp, argv[3], 1, 1, &f) == TCL_OK)
		{
		  soar_push_callback((soar_callback_agent) clientData, 
				     PRINT_CALLBACK,
				     (soar_callback_fn) Soar_PrintToFile,
				     (soar_callback_data) f,
				     (soar_callback_free_fn) NULL);
		}
	      else
		{
		  return TCL_ERROR;
		}
	    }
	}
      else if (string_match("-discard", argv[2]))
	{
	  soar_push_callback((soar_callback_agent) clientData, 
			     PRINT_CALLBACK,
			     (soar_callback_fn) Soar_DiscardPrint,
			     (soar_callback_data) NULL,
			     (soar_callback_free_fn) NULL);
	}
      else if (string_match("-append-to-result", argv[2]))
	{
	  soar_push_callback((soar_callback_agent) clientData, 
			     PRINT_CALLBACK,
			     (soar_callback_fn) Soar_AppendResult,
			     (soar_callback_data) NULL,
			     (soar_callback_free_fn) NULL);
	}
      else
	{
	  sprintf(interp->result,
		  "Unrecognized argument to %s %s: %s",
		  argv[0], argv[1], argv[2]);
	  return TCL_ERROR;      
	}
    }
  else if (string_match("-pop", argv[1]))
    {
      soar_pop_callback((soar_callback_agent) clientData,
			PRINT_CALLBACK);
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized argument to %s: %s",
	      argv[0], argv[1]);
      return TCL_ERROR;      
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * PfCmd --
 *
 *      This is the command procedure for the "pf" command, 
 *      which finds Soar productions by characteristic.
 *
 *      pf is a production finding facility.  It allows you to
 *      find productions that either test a particular LHS pattern
 *      or produce particular RHS preferences.
 *
 *      The syntax of the lhs-conditions or rhs-actions is exactly
 *      their syntax within SP's.  In addition, the symbol '*' may
 *      be used as a wildcard for an attribute or value.  Note that
 *      variable names do not have to match the specific names used
 *      in productions.
 * 
 *      Specifying -chunks means only chunks are searched (and 
 *      -nochunks means no chunks are searched).
 * 
 * Examples:
 *
 *      Find productions that test that some object gumby has 
 *      attribute ^alive t, and test an operator named foo:
 *        pf {(<s> ^gumby <gv> ^operator.name foo)(<gv> ^alive t)}
 *
 *      Find productions that propose foo:
 *        pf -rhs {(<x> ^operator <op> +)(<op> ^name foo)}
 *
 *      Find productions that test the attribute ^pokey:
 *        pf {(<x> ^pokey *)}
 *
 *      Find productions that test the operator foo
 *        pf {(<s> ^operator.name foo)}
 *
 * Syntax: pf [-rhs|-lhs|-chunks|-nochunks|-show-bindings] {clauses}
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints the productions found
 *
 *----------------------------------------------------------------------
 */

int PfCmd (ClientData clientData, 
	   Tcl_Interp * interp,
	   int argc, char *argv[])
{  
  int i;

  list *current_pf_list = NIL;

  bool lhs = TRUE;
  bool rhs = FALSE;
  bool show_bindings = FALSE;
  bool just_chunks = FALSE;
  bool no_chunks = FALSE;

  if (argc == 1)
    {
      interp->result = "No arguments given.\nUsage: pf [-rhs|-lhs|-chunks|-nochunks|-show-bindings] clauses";
      return TCL_ERROR;
    }

  for (i = 1; i < argc; i++)
    {
      if (string_match_up_to(argv[i], "-lhs", 2))
	{
	  lhs = TRUE;
	}
      else if (string_match_up_to(argv[i], "-rhs", 2))
	{
	  rhs = TRUE;
	  lhs = FALSE;
	}
      else if (string_match_up_to(argv[i], "-show-bindings", 2))
	{
	  show_bindings = TRUE;
	}
      else if (string_match_up_to(argv[i], "-chunks", 2))
	{
	  just_chunks = TRUE ;
	}
      else if (string_match_up_to(argv[i], "-nochunks", 2))
	{
	  no_chunks = TRUE ;
	}
      else
	{
	  break;
	}
    }

  if ((*argv[i] == '-') || (*argv[i] == '('))
    {
      if (argc > i + 1) 
	{
	  interp->result = "Too many arguments given.\nUsage: pf [-rhs|-lhs|-chunks|-nochunks|-show-bindings] clauses";
	  return TCL_ERROR;
	}
      if (lhs)
	{
	  ((agent *)clientData)->alternate_input_string = argv[i];
	  ((agent *)clientData)->alternate_input_suffix = ") ";
	  get_lexeme();
	  read_pattern_and_get_matching_productions (&current_pf_list,
						     show_bindings,
						     just_chunks, no_chunks);
	  current_agent(current_char) = ' ';
	}
      if (rhs)
	{
	  ((agent *)clientData)->alternate_input_string = argv[i];
	  ((agent *)clientData)->alternate_input_suffix = ") ";
	  get_lexeme();
	  read_rhs_pattern_and_get_matching_productions (&current_pf_list,
							 show_bindings,
							 just_chunks, 
							 no_chunks);
	  current_agent(current_char) = ' ';
	}
      if (current_pf_list == NIL) 
	{
	  print("No matches.\n");
	}

      free_list(current_pf_list);
    }
  else
    {
      sprintf(interp->result,
	      "Unknown argument to %s command: %s",
	      argv[0], argv[i]);
      return TCL_ERROR;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * PreferencesCmd --
 *
 *      This is the command procedure for the "preferences" command, 
 *      which prints all the preferences for the given slot.
 *
 *      This command prints all the preferences for the slot given 
 *      by the 'id' and 'attribute' arguments.  The optional 'detail' 
 *      argument must be one of the following (-none is the default):
 *
 *        -none     -- prints just the preferences themselves
 *        -names    -- also prints the names of the productions that 
 *                     generated them
 *        -timetags -- also prints the timetags of the wmes matched by 
 *                     the productions
 *        -wmes     -- prints the whole wmes, not just their timetags.
 *
 * Syntax:  preferences id attribute [detail]
 * Syntax:  preferences id attribute [ 0 | 1 | 2 | 3 ]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints preference information.
 *
 *----------------------------------------------------------------------
 */

int PreferencesCmd (ClientData clientData, 
		    Tcl_Interp * interp,
		    int argc, char *argv[])
{
  static char * too_few_args  = "Too few arguments.\nUsage: preferences id attribute [detail]";
  static char * too_many_args = "Too many arguments.\nUsage: preferences id attribute [detail]";

  Symbol *id, *attr;
  bool print_productions;
  wme_trace_type wtt;
  slot *s;
  preference *p;
  int i;
  
  if (argc < 3)
    {
      interp->result = too_few_args;
      return TCL_ERROR;
    }

  if (argc > 4)
    {
      interp->result = too_many_args;
      return TCL_ERROR;
    }

  /* --- read id --- */

  if (read_id_or_context_var_from_string(argv[1], &id)
      == TCL_ERROR)
    {
      sprintf(interp->result,
	      "Known id or context var expected in preferences command, not %s.",
	      argv[1]);
      return TCL_ERROR;
    }

  /* --- read attribute --- */

  get_lexeme_from_string(argv[2]);

  switch (current_agent(lexeme).type) {
  case SYM_CONSTANT_LEXEME:
    attr = find_sym_constant (current_agent(lexeme).string); 
    break;
  case INT_CONSTANT_LEXEME:
    attr = find_int_constant (current_agent(lexeme).int_val); 
    break;
  case FLOAT_CONSTANT_LEXEME:
    attr = find_float_constant (current_agent(lexeme).float_val); 
    break;
  case IDENTIFIER_LEXEME:
    attr = find_identifier (current_agent(lexeme).id_letter, 
			    current_agent(lexeme).id_number); 
    break;
  case VARIABLE_LEXEME:
    attr = read_identifier_or_context_variable();
    if (!attr) 
      return TCL_ERROR;
    break;
  default:
    sprintf (interp->result,
	     "Expected constant or identifier for attribute, not %s.",
		   argv[2]);
    return TCL_ERROR;
  }

  if (argc == 3)
    { /* --- no optional level indicator --- */
      print_productions = FALSE;
      wtt = NONE_WME_TRACE;
    }
  else
    { /* --- read the optional level indicator --- */

      if (   string_match_up_to(argv[3], "-none", 3)
	  || string_match(argv[3], "0"))
	{
	  print_productions = FALSE;
	  wtt = NONE_WME_TRACE;
	}
      else if (   string_match_up_to(argv[3], "-names", 3)
	       || string_match(argv[3], "1"))
	{
	  print_productions = TRUE;
	  wtt = NONE_WME_TRACE;
	}
      else if (   string_match_up_to(argv[3], "-timetags", 2)
	       || string_match(argv[3], "2"))
	{
	  print_productions = TRUE;
	  wtt = TIMETAG_WME_TRACE;
	}
      else if (   string_match_up_to(argv[3], "-wmes", 2)
	       || string_match(argv[3], "3"))
	{
	  print_productions = TRUE;
	  wtt = FULL_WME_TRACE;
	}
      else
	{
	  sprintf(interp->result,
		  "Unrecognized option to %s command: %s", 
		  argv[0], argv[2]);
	  return TCL_ERROR;
 
	}
    }

  /* --- print the preferences --- */

  s = find_slot (id, attr);
  if (!s) 
    {
      sprintf(interp->result,
	      "There are no preferences for %s ^%s.",
	      argv[1], argv[2]);
    return TCL_ERROR;
    }

  print_with_symbols ("Preferences for %y ^%y:\n", id, attr);

  for (i = 0; i < NUM_PREFERENCE_TYPES; i++)
    {
      if (s->preferences[i]) 
	{
	  print ("\n%ss:\n", preference_name[i]);
	  for (p = s->preferences[i]; p; p = p->next)
	    {
	      print_preference_and_source (p, print_productions, wtt);
	    }
	}
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * PrintCmd --
 *
 *      This is the command procedure for the "print" command, which 
 *      prints various Soar objects.
 * 
 *      The print command is used to print items from production 
 *      memory or working memory.  It can take several kinds of 
 *      arguments.
 *
 *      See also:  default-wme-depth
 *
 * Syntax:  print <stack>|<wmes>
 *          <stack> ::= -stack [-goal | -operator]*
 *          <wmes>  ::= [-depth n] [-internal] <arg>*
 *          <arg>   ::= production-name  (print that production)
 *          <arg>   ::= identifier       (id of the object to print)
 *          <arg>   ::= integer          (timetag of wme--the identifier 
 *                                       from the wme indicates the object 
 *                                       to be printed)
 *          <arg>   ::= <pattern>        (pattern--same as if you listed 
 *                                       as arguments the timetags of all 
 *                                       wmes matching the pattern)
 * 
 *        <pattern> ::= '{' '('  {identifier | '*'} 
 *                             ^ {attribute  | '*'} 
 *                               {value      | '*'} 
 *                               [+] ')' '}'
 *
 *      The optional [-depth n] argument overrides default-wme-depth.
 *      The optional [-internal] argument tells Soar to print things 
 *      in their internal form.  For productions, this means leaving 
 *      conditions in their reordered (rete net) form.  For wmes, this 
 *      means printing the individual wmes with their timetags, rather 
 *      than the objects.
 * 
 *      -depth 0 is meaningful only for integer and pattern arguments.  
 *      It causes only the matching wmes to be printed, instead of all 
 *      wmes whose id is an id in one of the matching wmes.
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints the selected objects.
 *
 *----------------------------------------------------------------------
 */

int PrintCmd (ClientData clientData, 
	      Tcl_Interp * interp,
	      int argc, char *argv[])
{
  static char * too_few_args  = "Too few arguments.\nUsage: print [-depth n] [-internal] arg*";

  bool internal;
  int depth;
  Symbol *id;
  wme *w;
  list *wmes;
  cons *c;
  int next_arg;

  internal = FALSE;
  depth = current_agent(default_wme_depth);   /* AGR 646 */

  if (argc == 1)
    {
      interp->result = too_few_args;
      return TCL_ERROR;
    }

  next_arg = 1;

  /* --- see if we have the -stack option --- */
  if (string_match("-stack", argv[next_arg])) 
    {
      bool print_goals;
      bool print_operators;

      /* Determine context items to print */

      if (argc == 2)                    /* No options given */
	{
	  print_goals     = TRUE;
	  print_operators = TRUE;
	}
      else
	{
	  int i;

	  print_goals     = FALSE;
	  print_operators = FALSE;

	  next_arg++;
	  for (i = next_arg; i < argc; i++)
	    {
	      if (string_match_up_to(argv[i], "-goal", 2))
		print_goals = TRUE;
	      else if (string_match_up_to(argv[i], "-operator", 2))
		print_operators      = TRUE;
	      else
		{
		  sprintf(interp->result,
			  "Unknown option passed to print -stack: (%s).\nUsage print -stack [-goal | -operator]*",
			  argv[next_arg]);
		  return TCL_ERROR;
		}
	    }
	}

      {
	Symbol *g;

	for (g = current_agent(top_goal); g != NIL; g = g->id.lower_goal) 
	  {
	    if (print_goals)
	      {
		print_stack_trace (g, g, FOR_STATES_TF, FALSE);
		print ("\n");
	      }
	    if (print_operators && g->id.operator_slot->wmes) 
	      {
		print_stack_trace (g->id.operator_slot->wmes->value,
				   g, FOR_OPERATORS_TF, FALSE);
		print ("\n");
	      }
	  }
      }

      return TCL_OK;      
    } /* End of string_match "-stack" */

  /* --- read optional -depth flag --- */
  if (string_match("-depth", argv[next_arg]))
    {
      if (argc == 2)
	{
	  interp->result = too_few_args;
	  return TCL_ERROR;
	}
      else
	{
	  if (Tcl_GetInt(interp, argv[++next_arg], &depth) != TCL_OK)
	    {
	      sprintf (interp->result,
		       "Integer expected after %s %s, not %s.",
		       argv[0], argv[1], argv[2]);
	      return TCL_ERROR;
	    }
	  else
	    {
	      next_arg++;
	    }
	}
    }

  /* --- read optional -internal flag --- */

  if (string_match("-internal", argv[next_arg]))
    {
      internal = TRUE;
      next_arg++;
    }

  /* --- repeat: read one arg and print it --- */
  while (next_arg < argc)
    {
      get_lexeme_from_string(argv[next_arg]);

      switch (current_agent(lexeme).type) {
      case SYM_CONSTANT_LEXEME:
	do_print_for_production_name (argv[next_arg],
				      internal);
	break;
      
      case INT_CONSTANT_LEXEME:
	for (w=current_agent(all_wmes_in_rete); w!=NIL; w=w->rete_next)
	  if (w->timetag == current_agent(lexeme).int_val) break;
	if (w) {
	  do_print_for_wme (w, depth, internal);
	} else {
	  sprintf(interp->result,
		  "No wme %ld in working memory", 
		  current_agent(lexeme).int_val);
	  return TCL_ERROR;
	}
	break;

      case IDENTIFIER_LEXEME:
      case VARIABLE_LEXEME:
	id = read_identifier_or_context_variable();
	if (id) do_print_for_identifier (id, depth, internal);
	break;
      
      case QUOTED_STRING_LEXEME:
	((agent *)clientData)->alternate_input_string = argv[next_arg];
	((agent *)clientData)->alternate_input_suffix = ") ";
	get_lexeme();
	wmes = read_pattern_and_get_matching_wmes ();
	current_agent(current_char) = ' ';
	for (c=wmes; c!=NIL; c=c->rest)
	  do_print_for_wme (c->first, depth, internal);
	free_list (wmes);
	break;

      default:
	sprintf(interp->result,
		"Unrecognized argument to %s command: %s",
		argv[0], argv[next_arg]);
	return TCL_ERROR;
      } /* end of switch statement */

      next_arg++;
  } /* end of while loop */
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * PwatchCmd --
 *
 *      This is the command procedure for the "pwatch" command
 *      which enables the tracing of production firing.
 *
 *      This command enables tracing of the firings and retractions 
 *      of individual productions.  (This mechanism is orthogonal to 
 *      the watch -firings mechanism.  See the "watch" command for more 
 *      information.  Pwatch, with no arguments, lists the productions 
 *      currently being traced.  With one or more production name 
 *      arguments, it enables tracing of those productions.  Tracing 
 *      persists until disabled by an unpwatch command, or until the 
 *      production is excised.
 *
 *      See also:  unpwatch, watch, excise
 *
 * Syntax:  pwatch production-name*
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Enables the tracing of the specified productions.
 *
 *----------------------------------------------------------------------
 */

int PwatchCmd (ClientData clientData, 
	       Tcl_Interp * interp,
	       int argc, char *argv[])
{
  if (argc == 1)
    {
      cons *c;

      for (c=current_agent(productions_being_traced); c!=NIL; c=c->rest)
	print_with_symbols (" %y\n", ((production *)(c->first))->name);  

      return TCL_OK;
    }
  else
    {
      int i;

      for (i = 1; i < argc; i++)
	{
	  production * prod;

	  if ((prod = name_to_production(argv[i])))
	    {
	      add_pwatch(prod);
	    }
	  else
	    {
	      sprintf(interp->result, "No production named %s", argv[i]);
	      return TCL_ERROR;
	    }
	}

      return TCL_OK;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * QuitCmd --
 *
 *      This is the command procedure for the "quit" command, 
 *      which exits Soar after cleaning up.
 * 
 * Syntax:  quit
 *
 * Results:
 *      Does not return.
 *
 * Side effects:
 *      Exits the Soar system.
 *
 *----------------------------------------------------------------------
 */

int QuitCmd (ClientData clientData, 
	     Tcl_Interp * interp,
	     int argc, char *argv[])
{
  static char cmd[] = "tcl-exit";
  if (argc > 1)
    {
      interp->result = "Too many arguments, should be: quit";
      return TCL_ERROR;
    }

  print ("Exiting Soar...\n");
  just_before_exit_soar();
  (void) Tcl_Eval(interp, cmd);
  return TCL_OK; /* Unreachable, but here to placate the compiler */
}

/*
 *----------------------------------------------------------------------
 *
 * RemoveWmeCmd --
 *
 *      This is the command procedure for the "remove-wme" command, 
 *      which removes a wme from Soar's memory.
 *
 *      WARNING: this command is inherently unstable and may have 
 *      weird side effects (possibly even including system crashes).  
 *      For example, the chunker can't backtrace through wmes created 
 *      via add-wme.  Removing input wmes or context/impasse wmes may 
 *      have unexpected side effects.  You've been warned.
 *
 *      See also:  add-wme
 *
 * Syntax:  remove-wme integer
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Removes the selected working memory element from Soar.
 *
 *----------------------------------------------------------------------
 */

int RemoveWmeCmd (ClientData clientData, 
		  Tcl_Interp * interp,
		  int argc, char *argv[])
{
  int num;

  if (argc == 1)
    {
      interp->result = "Too few arguments, should be: remove-wme integer";
      return TCL_ERROR;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments, should be: remove-wme integer";
      return TCL_ERROR;
    }
  
  if (Tcl_GetInt(interp, argv[1], &num) == TCL_OK)
    {
      wme *w, *w2;
      Symbol *id;
      slot *s;
      
      for (w=current_agent(all_wmes_in_rete); w!=NIL; w=w->rete_next)
	if (w->timetag == (unsigned long) num) 
	  break;

      if (!w) {
	sprintf (interp->result, "No wme %d in working memory\n", num);
	return TCL_ERROR;
      }

      id = w->id;

      /* --- remove w from whatever list of wmes it's on --- */
      for (w2=id->id.input_wmes; w2!=NIL; w2=w2->next)
	if (w==w2) break;
      if (w2) remove_from_dll (id->id.input_wmes, w, next, prev);
      for (w2=id->id.impasse_wmes; w2!=NIL; w2=w2->next)
	if (w==w2) break;
      if (w2) remove_from_dll (id->id.impasse_wmes, w, next, prev);
      for (s=id->id.slots; s!=NIL; s=s->next) {
	for (w2=s->wmes; w2!=NIL; w2=w2->next)
	  if (w==w2) break;
	if (w2) remove_from_dll (s->wmes, w, next, prev);    
	for (w2=s->acceptable_preference_wmes; w2!=NIL; w2=w2->next)
	  if (w==w2) break;
	if (w2) remove_from_dll (s->acceptable_preference_wmes, w, next, prev);
      }

      /* --- now remove w from working memory --- */
      remove_wme_from_wm (w);
      do_buffered_wm_and_ownership_changes();
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized argument to remove-wme command: %s",
	      argv[1]);
      return TCL_ERROR;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ReteNetCmd --
 *
 *      This is the command procedure for the "rete-net" command, 
 *      which saves and restores the state of the Rete network.
 *
 * Syntax:  rete-net option filename
 *            option ::= -save | -load
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Loads or saves the Rete network using the file "filename"
 *
 *----------------------------------------------------------------------
 */

int ReteNetCmd (ClientData clientData, 
		Tcl_Interp * interp,
		int argc, char *argv[])
{
  char * pipe_command;
  char * access_method;
  bool (*rete_net_op) (FILE *);
  bool using_compression_filter;

  Tcl_DString buffer;
  Tcl_DString command_line;
  char * fullname;
  FILE * f;
  bool result;
  bool loading;

  if (argc < 3)
    {
      interp->result = "Too few arguments.\nUsage: rete-net {-save | -load} filename.";
      return TCL_ERROR;
    }

  if (argc > 3)
    {
      interp->result = "Too many arguments.\nUsage: rete-net {-save | -load} filename.";
      return TCL_ERROR;
    }
  
  if (string_match("-save", argv[1]))
    {
      if (current_agent(all_productions_of_type)[JUSTIFICATION_PRODUCTION_TYPE]) 
	{
	  interp->result = "Can't save rete while there are justifications.";
	  return TCL_ERROR;
	}

      pipe_command  = "compress > ";
      access_method = "w";
      rete_net_op   = save_rete_net;        /* A function pointer */
      loading       = FALSE;
    }
  else if (string_match("-load", argv[1]))
    {
      int i;

      /* --- check for empty system --- */
      if (current_agent(all_wmes_in_rete)) {
	interp->result = "Can't load rete unless working memory is empty.";
	return TCL_ERROR;
      }

      for (i=0; i<NUM_PRODUCTION_TYPES; i++)
	if (current_agent(num_productions_of_type)[i]) 
	  {
	    interp->result = "Can't load rete unless production memory is empty.";
	    return TCL_ERROR;
	  }
  
      pipe_command  = "zcat ";
      access_method = "r";
      rete_net_op   = load_rete_net;        /* A function pointer */
      loading       = TRUE;
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized argument to rete-net command: %s",
	      argv[1]);
      return TCL_ERROR;
    }

  fullname = Tcl_TildeSubst(interp, argv[2], &buffer);
  if (fullname == NULL)
    {
      return TCL_ERROR;
    }
  
  /* cd to dir on top of stack ? */

  Tcl_DStringInit(&command_line);
  
  if (   (!(strcmp((char *) (argv[2] + strlen(argv[2]) - 2), ".Z")))
      || (!(strcmp((char *) (argv[2] + strlen(argv[2]) - 2), ".z"))))
    {
      /* The popen can succeed if given an non-existant file   */
      /* creating an unusable pipe.  So we check to see if the */
      /* file exists first, on a load action.                  */
      if (loading)
	{
	  if (!(f = fopen(fullname, access_method)))
	    {
	      /* --- error when opening the file --- */
	      Tcl_DStringFree(&command_line);
	      Tcl_DStringFree(&buffer);
	      sprintf (interp->result,
		       "rete-net: Error: unable to open '%s'",
		       argv[2]);
	      return TCL_ERROR;
	    }
	  else
	    {
	      fclose(f);
	    }
	}

      Tcl_DStringAppend(&command_line, pipe_command, strlen(pipe_command));
      Tcl_DStringAppend(&command_line, fullname, strlen(fullname));
      f = (FILE *) popen(Tcl_DStringValue(&command_line), access_method);
      using_compression_filter = TRUE;
    }
  else
    {
      Tcl_DStringAppend(&command_line, fullname, strlen(fullname));
      f = fopen(Tcl_DStringValue(&command_line), access_method);
      using_compression_filter = FALSE;
    }

  if (! f) 
    {
      /* --- error when opening the pipe or file --- */
      Tcl_DStringFree(&command_line);
      Tcl_DStringFree(&buffer);
      sprintf (interp->result,
	       "rete-net: Error: unable to open '%s'",
	       argv[2]);
      return TCL_ERROR;
    }

  result = (*rete_net_op)(f);

  if (using_compression_filter == TRUE)
    {
      pclose(f);
    }
  else
    {
      fclose(f);
    }

  Tcl_DStringFree(&command_line);
  Tcl_DStringFree(&buffer);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ScheduleCmd --
 *
 *	This is the command procedure for the "schedule" command, 
 *	which runs the Soar scheduler.
 *
 *      Schedule agents in a multi-agent soar setting.  If given, 
 *      "cycles" indicates how many cycles to run the scheduler.  
 *      If not given, the scheduler runs until interrupted with a 
 *      control-c or all agents have completed processing.
 *
 * Syntax:  schedule [cycles]
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Runs the Soar scheduler.  Terminates when the given number
 *      of cycles elapses or the user interrupts the scheduling
 *      process.
 *
 *----------------------------------------------------------------------
 */

int ScheduleCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  int cycles;
  
  if (argc > 2)
    {
      interp->result = "Too many arguments";
      return TCL_ERROR;
    }

  if (argc == 1)
    { /* No argument given, so we run schedule "forever" */
      cycles = -1;
    }
  else
    {
      if (Tcl_GetInt(interp, argv[1], &cycles) != TCL_OK)
	{
	  return TCL_ERROR;
	}
    }

  return schedule_agents(interp, cycles);
}

/*
 *----------------------------------------------------------------------
 *
 * SpCmd --
 *
 *      This is the command procedure for the "sp" command, which 
 *      defines a new Soar production.
 *
 *      This command adds a new production to the system.  (If 
 *      another production with the same name already exists, it is 
 *      excised.)  The optional flags are as follows: 
 *
 *        :o-support -- specifies that all the RHS actions are to 
 *                      be given o-support when the production fires
 *       :no-support -- specifies that all the RHS actions are only 
 *                      to be given i-support when the production fires
 *          :default -- specifies that this production is a default 
 *                      production (this matters for (excise-task) and 
 *                      (watch task))
 *            :chunk -- specifies that this production is a chunk (this 
 *                      matters for (learn trace))
 *
 * See also:  lhs-grammar, rhs-grammar
 *
 * Syntax: sp {rule-body}
 *
 *         rule-body := name 
 *                      ["documentation-string"] 
 *                      [ flag ]*
 *                      LHS
 *                      -->
 *                      RHS
 *
 *         flag  ::=  :o-support
 *         flag  ::=  :i-support
 *         flag  ::=  :default
 *         flag  ::=  :chunk
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Runs the agent for the indicated number of decision cycles.
 *
 *----------------------------------------------------------------------
 */

int SpCmd (ClientData clientData, 
	   Tcl_Interp * interp,
	   int argc, char *argv[])
{
  production *p;

  if (argc == 1)
    {
      interp->result = "Too few arguments.\nUsage: sp {rule}.";
      return TCL_ERROR;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments.\nUsage: sp {rule}.";
      return TCL_ERROR;
    }

  /* Use a callback instead? */
  ((agent *)clientData)->alternate_input_string = argv[1];
  ((agent *)clientData)->alternate_input_suffix = ") ";
  set_lexer_allow_ids (FALSE);
  get_lexeme();
  p = parse_production();
  set_lexer_allow_ids (TRUE);
  if (p) 
    {
      print("*");
      interp->result = "\n";	  
      return TCL_OK;
    }
  else
    {
      return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * StatsCmd --
 *
 *      This is the command procedure for the "stats" command, which 
 *      prints out internal statistical information.
 *
 * Syntax: stats [-system <stype> | -memory <mtype> | -rete <rtype>] 
 *         <stype> ::= -default-production-count
 *                     -user-production-count
 *                     -chunk-count
 *                     -justification-count
 *                     -all-productions-count
 *                     -dc-count
 *                     -ec-count
 *                     -ecs/dc
 *                     -firings-count
 *                     -firings/ec
 *                     -wme-change-count
 *                     -wme-addition-count
 *                     -wme-removal-count
 *                     -wme-count
 *                     -wme-avg-count
 *                     -wme-max-count
 *                     -total-time             |T
 *                     -ms/dc                  |T
 *                     -ms/ec                  |T
 *                     -ms/firing              |T
 *                     -ms/wme-change          |T
 *                     -match-time             |D
 *                     -ownership-time         |D
 *                     -chunking-time          |D
 *                     -o-support-time         |D
 *                     -instantiations-time    |D 
 *                     -other-pref-phase-time  |D
 *
 *         The items marked |T are available when Soar has been
 *         compiled with the NO_TIMING_STUFF flag NOT SET and 
 *         the items marked |D are available when Soar has been
 *         compiled with the DETAILED_TIMING_STATS flag SET.
 * 
 *         <mtype> ::= -total
 *                     -overhead
 *                     -strings
 *                     -hash-table
 *                     -pool [-total | pool-name [<aspect>]]
 *                     -misc
 *        <aspect> ::= -used                   |M
 *                     -free                   |M
 *                     -item-size
 *                     -total-bytes
 *
 *          The items marked |M are available only when Soar has
 *          been compiled with the MEMORY_POOL_STATS option.
 *
 *          <rtype> ::= -total-nodes
 *                      -dummy-top-nodes
 *                      -pos-nodes
 *                      -unhashed-pos-nodes
 *                      -neg-nodes
 *                      -unhashed-neg-nodes
 *                      -cn-nodes
 *                      -cn-partner-nodes
 *                      -production-nodes
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints the selected statistics
 *
 *----------------------------------------------------------------------
 */

int StatsCmd (ClientData clientData, 
	      Tcl_Interp * interp,
	      int argc, char *argv[])
{
  if (   (argc == 1) 
      || string_match_up_to("-system", argv[1], 2))
    {
      return parse_system_stats(interp, argc, argv);
    }
  else if (string_match_up_to("-memory", argv[1], 2))
    {
      return parse_memory_stats(interp, argc, argv);
    }
  else if (string_match_up_to("-rete", argv[1], 2))
    {
      return parse_rete_stats(interp, argc, argv);
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized argument to stats: %s",
	      argv[1]);
      return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * StopSoarCmd --
 *
 *      This is the command procedure for the "stop-soar" command, 
 *      halts the Soar agents.
 * 
 * Syntax:  stop-soar
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      All agents are halted.
 *
 *----------------------------------------------------------------------
 */

int StopSoarCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  if (argc > 1)
    {
      interp->result = "Too many arguments, should be: stop-soar";
      return TCL_ERROR;
    }

  control_c_handler(0);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UnPwatchCmd --
 *
 *      This is the command procedure for the "unpwatch" command
 *      which enables the tracing of production firing.
 *
 *      This command disables tracing of the firings and retractions 
 *      of individual productions.  (This mechanism is orthogonal to 
 *      the watch -firings mechanism.  See the "watch" command for more 
 *      information.  UnPwatch, with no arguments, disables all 
 *      active productions traces.  With one or more production name 
 *      arguments, it disables tracing of those productions.  
 *
 *      See also:  pwatch, watch
 *
 * Syntax:  unpwatch production-name*
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Disables the tracing of the specified productions.
 *
 *----------------------------------------------------------------------
 */

int UnPwatchCmd (ClientData clientData, 
	       Tcl_Interp * interp,
	       int argc, char *argv[])
{
  if (argc == 1)
    {
      cons *c;

      for (c=current_agent(productions_being_traced); c!=NIL; c=c->rest)
	{
	  production * prod;

	  prod = current_agent(productions_being_traced)->first;
	  remove_pwatch (prod);
	}

      return TCL_OK;
    }
  else
    {
      int i;
      
      for (i = 1; i < argc; i++)
	{
	  production * prod;

	  if ((prod = name_to_production(argv[i])))
	    {
	      remove_pwatch(prod);
	    }
	  else
	    {
	      sprintf(interp->result, "No production named %s", argv[i]);
	      return TCL_ERROR;
	    }
	}
      
      return TCL_OK;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * UserSelectCmd --
 *
 *      This is the command procedure for the "user-select" command, 
 *      which controls indifferent preference arbitration.
 *
 *      With no arguments, this command prints the current setting of 
 *      user-select.  With an argument, it sets user-select to the 
 *      given value.  This controls how Soar's decision procedure 
 *      chooses between multiple indifferent items:
 *        -first -- just choose the first one found (deterministically)
 *        -last -- choose the last one found (deterministically)
 *        -ask -- ask the user to choose one of the items
 *        -random -- choose one randomly
 *        -t -- synonymous with 'ask'
 *        -nil -- synonymous with 'random'
 *
 * Syntax:  user-select [-first | -ask | -last | -random | -t | -nil]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Modifies the setting of the user-select option.
 *
 *----------------------------------------------------------------------
 */

/* AGR 615  Adding the "last" option to this command was pretty simple
   and the changes are integrated into this entire function.  94.11.08 */

int UserSelectCmd (ClientData clientData, 
		   Tcl_Interp * interp,
		   int argc, char *argv[])
{
  if (argc == 1)
    {
      switch (current_agent(sysparams)[USER_SELECT_MODE_SYSPARAM]) 
	{
	case USER_SELECT_FIRST:  interp->result = "-first";  break;
	case USER_SELECT_LAST:   interp->result = "-last";   break;
	case USER_SELECT_ASK:    interp->result = "-ask";    break;
	case USER_SELECT_RANDOM: interp->result = "-random"; break;
	}
      return TCL_OK;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments, should be: user-select [-first | -ask | -random | -t | -nil]";
      return TCL_ERROR;
    }

  if (   (string_match_up_to(argv[1], "-ask", 2))
      || (string_match_up_to(argv[1], "-t", 2)))
    {
      set_sysparam (USER_SELECT_MODE_SYSPARAM, USER_SELECT_ASK);
    }
  else if (string_match_up_to(argv[1], "-first", 2))
    {
      set_sysparam (USER_SELECT_MODE_SYSPARAM, USER_SELECT_FIRST);
    }
  else if (string_match_up_to(argv[1], "-last", 2))
    {
      set_sysparam (USER_SELECT_MODE_SYSPARAM, USER_SELECT_LAST);
    }
  else if (   (string_match_up_to(argv[1], "-random", 2))
	   || (string_match_up_to(argv[1], "-nil", 2)))
    {
      set_sysparam (USER_SELECT_MODE_SYSPARAM, USER_SELECT_RANDOM);
    }
  else
    {
      sprintf(interp->result, 
	      "Unrecognized argument to user-select: %s", argv[1]);
      return TCL_ERROR;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * WarningsCmd --
 *
 *      This is the command procedure for the "warnings" command, 
 *      which enables/disables the printing of warning messages.
 *
 *      warnings -on enables the printing of warning messages.  This 
 *      is the default.  warnings -off turns off most warning messages.
 *      warnings prints an indication of whether warning messages 
 *      are enabled or not.
 *
 * Syntax:  warnings [-on | -off]
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Sets the warnings option.
 *
 *----------------------------------------------------------------------
 */

int WarningsCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  if (argc == 1)
    {
      sprintf(interp->result, "%s",
	      current_agent(sysparams)[PRINT_WARNINGS_SYSPARAM] 
	      ? "on" : "off");
      return TCL_OK;
    }

  if (argc > 2)
    {
      interp->result = "Too many arguments, should be: warnings [-on | -off]";
      return TCL_ERROR;
    }

  if (string_match_up_to(argv[1], "-on", 3))
    {
      set_sysparam (PRINT_WARNINGS_SYSPARAM, TRUE);
    }
  else if (string_match_up_to(argv[1], "-off", 3))
    {
    set_sysparam (PRINT_WARNINGS_SYSPARAM, FALSE);
    }
  else
    {
      sprintf(interp->result,
	      "Unrecognized option to warnings command: %s",
	      argv[1]);
      return TCL_ERROR;
    }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * WatchCmd --
 *
 *      This is the command procedure for the "watch" command, which 
 *      controls run-time tracing.
 *
 *      This command controls what information gets printed in the 
 *      run-time trace.  With no arguments, it just prints out the 
 *      current watch status.  The numeric arguments have roughly the 
 *      same semantics as in Soar 5; for details, see (help 
 *      watch-levels).  The various -keyword arguments are used to 
 *      modify the current watch settings.  For example, (watch 
 *      -context on) turns on the tracing of context slot decisions; 
 *      (watch -context off) turns it off again.  For information 
 *      about what each keyword does, see (help watch-keywords).
 *
 *      The following keyword arguments may be given to the 'watch' 
 *      command:
 *
 *      -context -- controls whether context slot decisions are printed
 *      -phases -- controls whether phase names are printed
 *      -firings [default|user|chunk|nonchunk|all] -- controls which 
 *               production firings and retractions are printed.  The 
 *               optional argument (which defaults to 'all') specifies 
 *               which types of productions the 'on/off' argument 
 *               refers to.
 *      -firings-wmes -- controls the level of detail given about the 
 *               wmes matched by productions whose firings and 
 *               retractions are being traced.  Level 0 means no 
 *               information about the wmes is printed.  Level 1 means 
 *               the wme timetags are printed.  Level 2 means the whole 
 *               wmes are printed.
 *      -firings-preferences -- controls whether the preferences 
 *               generated by the traced productions are printed when 
 *               those productions fire or retract.  When a production 
 *               fires, all the preferences it generates are printed.  
 *               When it retracts, only the ones being removed from
 *               preference memory are printed (i.e., the i-supported 
 *               ones).
 *      -wmes -- controls whether changes to working memory are printed
 *      -chunk-names -- controls whether names of newly built chunks 
 *               are printed
 *      -justification-names -- ditto, only for justifications 
 *               (internal chunks)
 *      -chunks -- controls whether newly built chunks are printed
 *      -justifications -- ditto, only for justifications
 *      -backtracing -- controls whether backtracing information is 
 *               printed
 *
 *      The watch levels (-1, 0, 0.5, 1, 1.5, 2, 3) in Soar 6 have 
 *      roughly the same semantics as in Soar 5.  The table below gives 
 *      the corresponding keyword parameter settings for each level.
 *
 *                   Watch Level:    -1    0  0.5    1  1.5    2    3
 *                                  ---- ---- ---- ---- ---- ---- ----
 *         -context                 off   on   on   on   on   on   on
 *         -phases                  off  off   on   on   on   on   on
 *         -firings nonchunk        off  off   on   on   on   on   on
 *           -firings-wmes            0    0    0    1    2    2    2
 *           -firings-preferences   off  off  off  off  off  off   on
 *         -wmes                    off  off  off  off  off   on   on
 *
 *      (watch task) is also provided in Soar 6 for Soar 5 compatibility, 
 *      and is equivalent to (watch -firings default off).
 *
 *      Learn noprint/print/full-print and notrace/trace/full-trace also 
 *      translate into keyword parameter settings, as shown in the tables 
 *      below:
 *
 *                 Learn:   noprint  print  full-print
 *                          -------  -----  ----------
 *         -chunk-names        off     on          on
 *         -chunks             off    off          on
 *
 *                 Learn:   notrace  trace  full-trace
 *                          -------  -----  ----------
 *         -firings chunk      off     on          on
 *         -backtracing        off    off          on
 *
 *      These keyword arguments are provided so you can have more 
 *      complete control over what gets traced.  The numeric arguments 
 *      to 'watch', as well as the 'print' and 'trace' arguments to 
 *      'learn', provide simple ways of getting some useful watch 
 *      settings.  See (help watch-levels) for details.
 *
 *  See also:  learn, pwatch
 *
 * Syntax: watch arg*
 *           arg  ::=  -1 | 0 | 0.5 | 1 | 1.5 | 2 | 3 | task
 *           arg  ::=  -context {on|off}
 *           arg  ::=  -phases {on|off}
 *           arg  ::=  -firings [default|user|chunk|nonchunk|all] {on|off}
 *           arg  ::=     -firings-wmes {0|1|2}
 *           arg  ::=     -firings-preferences {on|off}
 *           arg  ::=  -wmes {on|off}
 *           arg  ::=  -chunk-names {on|off}
 *           arg  ::=  -justification-names {on|off}
 *           arg  ::=  -chunks {on|off}
 *           arg  ::=  -justifications {on|off}
 *           arg  ::=  -backtracing {on|off}
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Runs the agent for the indicated number of decision cycles.
 *
 *----------------------------------------------------------------------
 */

int WatchCmd (ClientData clientData, 
	      Tcl_Interp * interp,
	      int argc, char *argv[])
{
  if (argc == 1)
    {
      print_current_watch_settings();
    }
  
  {
    int i;

    for (i = 1; i < argc; i++)
      {
	if (string_match("-context", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_CONTEXT_DECISIONS_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-phases", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_PHASES_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-wmes", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_WM_CHANGES_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-chunk-names", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_CHUNK_NAMES_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-justification-names", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_JUSTIFICATION_NAMES_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-chunks", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_CHUNKS_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-justifications", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_JUSTIFICATIONS_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-backtracing", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_BACKTRACING_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-firings-preferences", argv[i]))
	  {
	    if (set_watch_setting (interp,
				   TRACE_FIRINGS_PREFERENCES_SYSPARAM,
				   argv[i],
				   argv[++i]) 
		!= TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-task", argv[i]))
	  {
	    set_sysparam (TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, 
			  FALSE);	    
	  }
	else if (string_match("-firings", argv[i]))
	  {
	    bool types[NUM_PRODUCTION_TYPES];
	    int t;

	    i++;

	    if (string_match("default", argv[i]))
	      {
		for (t = 0; t < NUM_PRODUCTION_TYPES; t++) 
		  {
		    types[t] = FALSE;
		  }
		types[DEFAULT_PRODUCTION_TYPE] = TRUE;		
		i++;
	      }
	    else if (string_match("user", argv[i]))
	      {
		for (t = 0; t < NUM_PRODUCTION_TYPES; t++) 
		  {
		    types[t] = FALSE;
		  }
		types[USER_PRODUCTION_TYPE] = TRUE;		
		i++;
	      }
	    else if (string_match("chunk", argv[i]))
	      {
		for (t = 0; t < NUM_PRODUCTION_TYPES; t++) 
		  {
		    types[t] = FALSE;
		  }
		types[CHUNK_PRODUCTION_TYPE] = TRUE;		
		i++;
	      }
	    else if (string_match("nonchunk", argv[i]))
	      {
		for (t = 0; t < NUM_PRODUCTION_TYPES; t++) 
		  {
		    types[t] = FALSE;
		  }
		types[DEFAULT_PRODUCTION_TYPE] = TRUE;		
		types[USER_PRODUCTION_TYPE]    = TRUE;		
		i++;
	      }
	    else /* argv[i] may be "all" or nothing */
	      {
		for (t = 0; t < NUM_PRODUCTION_TYPES; t++) 
		  {
		    types[t] = TRUE;
		  }
		if (string_match("all", argv[i]))
		  {
		    i++;
		  }
	      }

	    /* --- read on/off flag --- */

	    if (string_match("on", argv[i]))
	      {
		for (t = 0; t < NUM_PRODUCTION_TYPES; t++)
		  if (types[t])
		    set_sysparam (TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM+t,
				  TRUE);
	      }
	    else if (string_match("off", argv[i]))
	      {
		for (t = 0; t < NUM_PRODUCTION_TYPES; t++)
		  if (types[t])
		    set_sysparam (TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM+t,
				  FALSE);
	      }
	    else
	      {
		sprintf(interp->result,
			"Unrecognized option to watch parameter %s: %s",
			argv[i-1], argv[i]);
		return TCL_ERROR;
	      }
          } /* end of "-firings" */
	else if (string_match("-firings-wmes", argv[i]))
	  {
	    i++;

	    if (string_match("0", argv[i]))
	      {
		set_sysparam (TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,
			      NONE_WME_TRACE);
	      }
	    else if (string_match("1", argv[i]))
	      {
		set_sysparam (TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,
			      TIMETAG_WME_TRACE);
	      }
	    else if (string_match("2", argv[i]))
	      {
		set_sysparam (TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,
			      FULL_WME_TRACE);
	      }
	    else
	      {
		sprintf(interp->result,
			"Unexpected option to watch -firings-wmes: %s",
			argv[i]);
		return TCL_ERROR;
	      }
	  }
	else if (string_match("-1", argv[i]))
	  {
	    set_sysparam(TRACE_CONTEXT_DECISIONS_SYSPARAM,        FALSE);
	    set_sysparam(TRACE_PHASES_SYSPARAM,                   FALSE);
	    set_sysparam(TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, FALSE);
	    set_sysparam(TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM,    FALSE);
	    set_sysparam(TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,   NONE_WME_TRACE);
	    set_sysparam(TRACE_FIRINGS_PREFERENCES_SYSPARAM,      FALSE);
	    set_sysparam(TRACE_WM_CHANGES_SYSPARAM,               FALSE);
	  }
	else if (string_match("0", argv[i]))
	  {
	    set_sysparam(TRACE_CONTEXT_DECISIONS_SYSPARAM,        TRUE);
	    set_sysparam(TRACE_PHASES_SYSPARAM,                   FALSE);
	    set_sysparam(TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, FALSE);
	    set_sysparam(TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM,    FALSE);
	    set_sysparam(TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,   NONE_WME_TRACE);
	    set_sysparam(TRACE_FIRINGS_PREFERENCES_SYSPARAM,      FALSE);
	    set_sysparam(TRACE_WM_CHANGES_SYSPARAM,               FALSE);
	  }
	else if (string_match("0.5", argv[i]))
	  {
	    set_sysparam(TRACE_CONTEXT_DECISIONS_SYSPARAM,        TRUE);
	    set_sysparam(TRACE_PHASES_SYSPARAM,                   TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM,    TRUE);
	    set_sysparam(TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,   NONE_WME_TRACE);
	    set_sysparam(TRACE_FIRINGS_PREFERENCES_SYSPARAM,      FALSE);
	    set_sysparam(TRACE_WM_CHANGES_SYSPARAM,               FALSE);
	  }
	else if (string_match("1", argv[i]))
	  {
	    set_sysparam(TRACE_CONTEXT_DECISIONS_SYSPARAM,        TRUE);
	    set_sysparam(TRACE_PHASES_SYSPARAM,                   TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM,    TRUE);
	    set_sysparam(TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,   TIMETAG_WME_TRACE);
	    set_sysparam(TRACE_FIRINGS_PREFERENCES_SYSPARAM,      FALSE);
	    set_sysparam(TRACE_WM_CHANGES_SYSPARAM,               FALSE);
	  }
	else if (string_match("1.5", argv[i]))
	  {
	    set_sysparam(TRACE_CONTEXT_DECISIONS_SYSPARAM,        TRUE);
	    set_sysparam(TRACE_PHASES_SYSPARAM,                   TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM,    TRUE);
	    set_sysparam(TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,   FULL_WME_TRACE);
	    set_sysparam(TRACE_FIRINGS_PREFERENCES_SYSPARAM,      FALSE);
	    set_sysparam(TRACE_WM_CHANGES_SYSPARAM,               FALSE);
	  }
	else if (string_match("2", argv[i]))
	  {
	    set_sysparam(TRACE_CONTEXT_DECISIONS_SYSPARAM,        TRUE);
	    set_sysparam(TRACE_PHASES_SYSPARAM,                   TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM,    TRUE);
	    set_sysparam(TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,   FULL_WME_TRACE);
	    set_sysparam(TRACE_FIRINGS_PREFERENCES_SYSPARAM,      FALSE);
	    set_sysparam(TRACE_WM_CHANGES_SYSPARAM,               TRUE);
	  }
	else if (string_match("3", argv[i]))
	  {
	    set_sysparam(TRACE_CONTEXT_DECISIONS_SYSPARAM,        TRUE);
	    set_sysparam(TRACE_PHASES_SYSPARAM,                   TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_DEFAULT_PRODS_SYSPARAM, TRUE);
	    set_sysparam(TRACE_FIRINGS_OF_USER_PRODS_SYSPARAM,    TRUE);
	    set_sysparam(TRACE_FIRINGS_WME_TRACE_TYPE_SYSPARAM,   FULL_WME_TRACE);
	    set_sysparam(TRACE_FIRINGS_PREFERENCES_SYSPARAM,      TRUE);
	    set_sysparam(TRACE_WM_CHANGES_SYSPARAM,               TRUE);
	  }
	else
	  {
	    sprintf(interp->result,
		    "Unrecognized argument to watch command: %s",
		    argv[i]);
	    return TCL_ERROR;
	  }
      }
  }

  return TCL_OK;
}

void Soar_InstallCommands (agent * the_agent)
{
  install_tcl_soar_cmd(the_agent, "add-wme",             AddWmeCmd);
  install_tcl_soar_cmd(the_agent, "agent-go",            AgentGoCmd);
  install_tcl_soar_cmd(the_agent, "attribute-preferences-mode", AttributePreferencesModeCmd);
  install_tcl_soar_cmd(the_agent, "chunk-free-problem-spaces", CFPSCmd);
  install_tcl_soar_cmd(the_agent, "chunky-problem-spaces", CPSCmd);
  install_tcl_soar_cmd(the_agent, "default-wme-depth",   DefWmeDepthCmd);
  install_tcl_soar_cmd(the_agent, "echo",                EchoCmd);
  install_tcl_soar_cmd(the_agent, "excise",              ExciseCmd);
  install_tcl_soar_cmd(the_agent, "explain",             ExplainCmd);
  install_tcl_soar_cmd(the_agent, "firing-counts",       FiringCountsCmd);
  install_tcl_soar_cmd(the_agent, "format-watch",        FormatWatchCmd); 
  install_tcl_soar_cmd(the_agent, "go",                  GoCmd);
  install_tcl_soar_cmd(the_agent, "io",                  IOCmd);
  install_tcl_soar_cmd(the_agent, "init-soar",           InitSoarCmd);
  install_tcl_soar_cmd(the_agent, "input-period",        InputPeriodCmd);
  install_tcl_soar_cmd(the_agent, "internal-symbols",    InternalSymbolsCmd);  
  install_tcl_soar_cmd(the_agent, "learn",               LearnCmd);
  install_tcl_soar_cmd(the_agent, "list-productions",    ListProdCmd);
  install_tcl_soar_cmd(the_agent, "log",                 LogCmd);
  install_tcl_soar_cmd(the_agent, "matches",             MatchesCmd);
  install_tcl_soar_cmd(the_agent, "max-chunks",          MaxChunksCmd);
  install_tcl_soar_cmd(the_agent, "max-elaborations",    MaxElaborationsCmd);
  install_tcl_soar_cmd(the_agent, "memories",            MemoriesCmd);
  install_tcl_soar_cmd(the_agent, "monitor",             MonitorCmd);
  install_tcl_soar_cmd(the_agent, "multi-attribute",     MultiAttrCmd);
  install_tcl_soar_cmd(the_agent, "o-support-mode",      OSupportModeCmd);
  install_tcl_soar_cmd(the_agent, "output-strings-destination", OutputStringsDestCmd);
  install_tcl_soar_cmd(the_agent, "pf",                  PfCmd);
  install_tcl_soar_cmd(the_agent, "preferences",         PreferencesCmd);
  install_tcl_soar_cmd(the_agent, "print",               PrintCmd);
  install_tcl_soar_cmd(the_agent, "pwatch",              PwatchCmd);
  install_tcl_soar_cmd(the_agent, "quit",                QuitCmd);
  install_tcl_soar_cmd(the_agent, "remove-wme",          RemoveWmeCmd);
  install_tcl_soar_cmd(the_agent, "rete-net",            ReteNetCmd);
  install_tcl_soar_cmd(the_agent, "schedule",            ScheduleCmd);
  install_tcl_soar_cmd(the_agent, "sp",                  SpCmd);
  install_tcl_soar_cmd(the_agent, "stats",               StatsCmd);
  install_tcl_soar_cmd(the_agent, "stop-soar",           StopSoarCmd);
  install_tcl_soar_cmd(the_agent, "unpwatch",            UnPwatchCmd);
  install_tcl_soar_cmd(the_agent, "user-select",         UserSelectCmd);
  install_tcl_soar_cmd(the_agent, "warnings",            WarningsCmd);
  install_tcl_soar_cmd(the_agent, "watch",               WatchCmd);
}
