/* 
 * main.c --
 *
 *	Main program for Tcl shells and other Tcl-based applications.
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 * Copyright (c) 1995 Carnegie Mellon University, University of Michigan,
 *                    University of Southern California/Information Sciences
 *                    Institute.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * The Soar consortium proclaims this software is in the public domain, and
 * is made available AS IS.  Carnegie Mellon University, The University of 
 * Michigan, and The University of Southern California/Information Sciences 
 * Institute make no warranties about the software or its performance, implied
 * or otherwise.
 */

#include <errno.h>
#include <stdio.h>
#include <soar.h>

/*
 * Declarations for various library procedures and variables (don't want
 * to include tkInt.h or tkPort.h here, because people might copy this
 * file out of the Tk source directory to make their own modified versions).
 * Note:  "exit" should really be declared here, but there's no way to
 * declare it without causing conflicts with other definitions elsewher
 * on some systems, so it's better just to leave it out.
 */

extern int		isatty _ANSI_ARGS_((int fd));
extern int		read _ANSI_ARGS_((int fd, char *buf, size_t size));

/*
 * Global variables used by the main program:
 */

RegisteredInterp * mainInterp = NULL;
                                /* The current active top-level interp. */
static Tcl_Interp *interp;	/* Interpreter for this application. */
static Tcl_DString command;	/* Used to assemble lines of terminal input
				 * into Tcl commands. */
static int tty;			/* Non-zero means standard input is a
				 * terminal-like device.  Zero means it's
				 * a file. */
static char errorExitCmd[] = "exit 1";
static char * process_name;     /* Name used to start this process */

/*
 * Command-line options:
 */

static option_table options;    /* Table of interpreter creation options */

static Soar_ArgvInfo argTable[] = {
    {"-agent",	  SOAR_ARGV_AGENT,    (char *) NULL, (char *) 
       &options.names, "Define a new agent"},
    {"-path",SOAR_ARGV_STRING,   (char *) NULL, (char *) 
       &options.path, "Directory path to scan when defining interps"},
#ifdef USE_TK
    {"-display",  SOAR_ARGV_STRING,   (char *) NULL, (char *) 
       &options.display, "Display to use"},
#endif
    {"-file",     SOAR_ARGV_STRING,   (char *) NULL, (char *) 
       &options.fileName, "File from which to read commands"},
#ifdef USE_TK
    {"-geometry", SOAR_ARGV_STRING,   (char *) NULL, (char *) 
       &options.geometry, "Initial geometry for window"},
#endif
    {"-help",	  SOAR_ARGV_HELP,     (char *) NULL, (char *) 
       NULL, "Print summary of command-line options and abort"},
#ifdef USE_TK
    {"-noTk",     SOAR_ARGV_CONSTANT, (char *)    0, (char *) 
       &options.tk_enabled, "Disable use of Tk"},
    {"-sync",     SOAR_ARGV_CONSTANT, (char *)    1, (char *) 
       &options.synchronize, "Use synchronous mode for display server"},
#endif
    {"-tclsh",	  SOAR_ARGV_TCLSH,    (char *) NULL, (char *) 
       &options.names, "Define a new Tcl shell"},
    {"-verbose",  SOAR_ARGV_CONSTANT, (char *)    1, (char *) 
       &options.verbose, "Print detailed information on option processing"},
#ifdef USE_TK
    {"-wish",	  SOAR_ARGV_WISH,     (char *) NULL, (char *) 
       &options.names, "Define a new Tcl/Tk shell"},
#endif
    {NULL,	  SOAR_ARGV_END,      (char *) NULL, (char *) 
       NULL, (char *) NULL}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

#ifndef USE_TK
static void Tcl_MainLoop _ANSI_ARGS_((void));
#endif

static void InitOptions _ANSI_ARGS_((option_table * options));
static void PrintOptions _ANSI_ARGS_((char ** argv, option_table * options,
				      int interp_type, int options_done));
static void Prompt _ANSI_ARGS_((Tcl_Interp *interp, int partial));

#ifdef USE_TK
static void StdinProc _ANSI_ARGS_((ClientData clientData, int mask));
#endif

/*
 *----------------------------------------------------------------------
 *
 * Soar_Main --
 *
 *	Main program for soar, soartk, and most other Soar-based
 *      applications.
 *
 * Results:
 *	None. This procedure never returns (it exits the process when
 *	it's done).
 *
 * Side effects:
 *	This procedure initializes the Soar world and then starts
 *	interpreting commands;  almost anything could happen, depending
 *	on the script being interpreted.
 *
 *----------------------------------------------------------------------
 */

void
Soar_Main(argc, argv)
    int argc;				/* Number of arguments. */
    char **argv;			/* Array of argument strings. */
{
    char *msg;                          /* Error message */
    int code;                           /* Tcl interp return code */
    Tcl_Interp * dummyInterp;

    process_name = argv[0];
    tty = isatty(0);
    tcl_RcFileName = "~/.soarrc";

    /*
     * Init Soar subsystem.
     */

    init_soar();

    /*
     * Create a dummy interpreter to hold the error messages generated
     * while parsing the command line to define interpreters.  We do it
     * this way since the function Soar_DefineInterps could be invoked
     * interactively from a Tcl shell as well as the command line. 
     */

    dummyInterp = Tcl_CreateInterp();
    if (Soar_DefineInterps(&argc, argv, dummyInterp) != TCL_OK)
      {
	fprintf(stderr, dummyInterp->result);
	exit(1);
      }
    Tcl_DeleteInterp(dummyInterp);

    /* 
     * Add all unrecognized argv switches to registered interpreters 
     */
    
    Soar_RegisterArgv(argc, argv, NULL);

    /*
     * Select the most recently-defined interp as the one to start
     * with at the command line.  This will set interp, mainInterp,
     * and soar_agent global variables as a side-effect.
     */

    Soar_SelectGlobalInterp(registry->name);
    
    /*
     * Invoke the script specified on the command line, if any.
     */

    if (options.fileName != NULL) {
	code = Tcl_VarEval(interp, "source ", options.fileName, (char *) NULL);
	if (code != TCL_OK) {
	    goto error;
	}
	tty = 0;
    } else {
	/*
         * Evaluate the "rc" file, if one has been specified, set up 
         * an event handler for standard input, and print a prompt if 
         * the input device is a terminal.
	 */

	if (tcl_RcFileName != NULL) {
	    Tcl_DString buffer;
	    char *fullName;
	    FILE *f;
    
	    fullName = Tcl_TildeSubst(interp, tcl_RcFileName, &buffer);
	    if (fullName == NULL) {
		fprintf(stderr, "%s\n", interp->result);
	    } else {
		f = fopen(fullName, "r");
		if (f != NULL) {
		    code = Tcl_EvalFile(interp, fullName);
		    if (code != TCL_OK) {
			fprintf(stderr, "%s\n", interp->result);
		    }
		    fclose(f);
		}
	    }
	    Tcl_DStringFree(&buffer);
	}
#ifdef USE_TK
	Tk_CreateFileHandler(0, TK_READABLE, StdinProc, (ClientData) 0);
	if (tty) {
	    Prompt(interp, 0);
	}
#endif
    }
    fflush(stdout);
    Tcl_DStringInit(&command);
    Tcl_ResetResult(interp);

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */

#ifdef USE_TK    
    tk_NumMainWindows++;
    Tk_MainLoop();
#else
    Tcl_MainLoop();
#endif

    /*
     * Don't exit directly, but rather invoke the Tcl "exit" command.
     * This gives the application the opportunity to redefine "exit"
     * to do additional cleanup.
     */

    Tcl_Eval(interp, "exit");
    exit(1);

error:
    msg = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (msg == NULL) {
	msg = interp->result;
    }
    fprintf(stderr, "%s\n", msg);
    Tcl_Eval(interp, errorExitCmd);
    return;			/* Needed only to prevent compiler warnings. */
}

/*
 *----------------------------------------------------------------------
 *
 * StdinProc --
 *
 *	This procedure is invoked by the event dispatcher whenever
 *	standard input becomes readable.  It grabs the next line of
 *	input characters, adds them to a command being assembled, and
 *	executes the command if it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Could be almost arbitrary, depending on the command that's
 *	typed.
 *
 *----------------------------------------------------------------------
 */

#ifdef USE_TK
    /* ARGSUSED */
static void
StdinProc(clientData, mask)
    ClientData clientData;		/* Not used. */
    int mask;				/* Not used. */
{
#define BUFFER_SIZE 4000
    char input[BUFFER_SIZE+1];
    static int gotPartial = 0;
    char *cmd;
    int code, count;

    count = read(fileno(stdin), input, BUFFER_SIZE);
    if (count <= 0) {
	if (!gotPartial) {
	    if (tty) {
		Tcl_Eval(interp, "exit");
		exit(1);
	    } else {
		Tk_DeleteFileHandler(0);
	    }
	    return;
	} else {
	    count = 0;
	}
    }
    cmd = Tcl_DStringAppend(&command, input, count);
    if (count != 0) {
	if ((input[count-1] != '\n') && (input[count-1] != ';')) {
	    gotPartial = 1;
	    goto prompt;
	}
	if (!Tcl_CommandComplete(cmd)) {
	    gotPartial = 1;
	    goto prompt;
	}
    }
    gotPartial = 0;

    /*
     * Disable the stdin file handler while evaluating the command;
     * otherwise if the command re-enters the event loop we might
     * process commands from stdin before the current command is
     * finished.  Among other things, this will trash the text of the
     * command being evaluated.
     */

    Tk_CreateFileHandler(0, 0, StdinProc, (ClientData) 0);

    if (mainInterp->agentPtr != NULL)
      {
	Soar_Log(soar_agent, cmd);
      }

    code = Tcl_RecordAndEval(interp, cmd, TCL_EVAL_GLOBAL);
    Tk_CreateFileHandler(0, TK_READABLE, StdinProc, (ClientData) 0);
    Tcl_DStringFree(&command);
    if (*interp->result != 0) {
	if ((code != TCL_OK) || (tty)) {
	  /* Previous use of printf below (now is puts) core   */
          /* dumped under Solaris 2.3 if result was very long. */
	  puts(interp->result);
	  if (mainInterp->agentPtr != NULL)
	    {
	      Soar_Log(soar_agent, interp->result);
	      Soar_Log(soar_agent, "\n");
	    }
	}
    }

    /*
     * Output a prompt.
     */

    prompt:
    if (tty) {
	Prompt(interp, gotPartial);
    }
    Tcl_ResetResult(interp);
}
#endif

/*
 *----------------------------------------------------------------------
 *
 * Prompt --
 *
 *	Issue a prompt on standard output, or invoke a script
 *	to issue the prompt.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A prompt gets output, and a Tcl script may be evaluated
 *	in interp.
 *
 *----------------------------------------------------------------------
 */

static void
Prompt(interp, partial)
    Tcl_Interp *interp;			/* Interpreter to use for prompting. */
    int partial;			/* Non-zero means there already
					 * exists a partial command, so use
					 * the secondary prompt. */
{
    char *promptCmd;
    int code;

    promptCmd = Tcl_GetVar(interp,
			   partial ? "tcl_prompt2" : "tcl_prompt1", 
			   TCL_GLOBAL_ONLY);
    if (promptCmd == NULL) 
      {
      defaultPrompt:
	if (!partial) 
	  {
	    fputs("% ", stdout);
	    if (mainInterp->agentPtr != NULL)
	      {
		Soar_Log(soar_agent, "% ");
	      }
	  }
      } 
    else 
      {
	code = Tcl_Eval(interp, promptCmd);
	if (code != TCL_OK) {
	  Tcl_AddErrorInfo(interp,
			   "\n    (script that generates prompt)");
	  fprintf(stderr, "%s\n", interp->result);
	  goto defaultPrompt;
	}
      }

    fflush(stdout);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SelectGlobalInterp --
 *
 *	Select an interpreter to use at the top level.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Changes the global interpreter.
 *
 *----------------------------------------------------------------------
 */

void 
Soar_SelectGlobalInterp (name)
    char * name;                /* Name of interpreter to use at the
				 * top-level. */
{
  RegisteredInterp *riPtr;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (!strcmp(riPtr->name, name)) {
      mainInterp = riPtr;
      interp     = riPtr->interp;
      soar_agent = riPtr->agentPtr;
    }
  }
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_SelectGlobalInterpByInterp --
 *
 *	Select an interpreter to use at the top level.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Changes the global interpreter.
 *
 *----------------------------------------------------------------------
 */

void 
Soar_SelectGlobalInterpByInterp (interp)
    Tcl_Interp * interp;        /* interpreter to use at the top-level. */
{
  RegisteredInterp *riPtr;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (riPtr->interp == interp) {
      mainInterp = riPtr;
      interp     = riPtr->interp;
      soar_agent = riPtr->agentPtr;
    }
  }
}

/*
 *----------------------------------------------------------------------
 *
 * InitOptions --
 *
 *	Initialize the option table used to guide the creation of
 *      new interpreters from specified command line arguments.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Initializes table of interpreter options.
 *
 *----------------------------------------------------------------------
 */

static void 
InitOptions (options)
     option_table * options;
{
#ifdef USE_TK
  options->tk_enabled  = 1;
#else
  options->tk_enabled  = 0;
#endif

  options->synchronize = 0;
  options->verbose     = 0;

  options->path   = NULL;
  options->display     = NULL;
  options->fileName    = NULL;
  options->geometry    = NULL;

  options->names.start_argv = 0;
  options->names.end_argv   = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * PrintOptions --
 *
 *	Print the options defined for the current interpreter
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Prints option information to stdout.
 *
 *----------------------------------------------------------------------
 */

static void 
PrintOptions (argv, options, interp_type, options_done)
     char ** argv;                    /* Command line options */
     option_table * options;          /* Parsed options */
     int interp_type;                 /* Interpreter type */
     int options_done;                /* Argv parsed completely? */
{
  int i; 

  printf("\nOption Settings:\n");

  printf("-noTk:      %s\n", options->tk_enabled  ? "no" : "yes");
  printf("-sync:      %s\n", options->synchronize ? "yes" : "no");
  printf("-verbose:   %s\n", options->verbose     ? "yes" : "no");

  printf("-path:     \"%s\"\n", options->path);
  printf("-display:  \"%s\"\n", options->display);
  printf("-file:     \"%s\"\n", options->fileName);
  printf("-geometry: \"%s\"\n", options->geometry);

  switch (interp_type) {
  case SOAR_ARGV_TCLSH:
    printf("Tclsh List: ");
    break;
  case SOAR_ARGV_WISH:
    printf("Wish List:  ");
    break;
  case SOAR_ARGV_AGENT:
    printf("Agent List: ");
    break;
  default:
    break;
  }

  if (options->names.start_argv != 0)
    {
      for (i=options->names.start_argv; i < options->names.end_argv; i++) {
	printf ("%s, ", argv[i]);
      }
      printf ("%s\n", argv[i]);
    }
  else
    {
      printf("\n");
    }

  printf("\nOptions Done?: %s\n", options_done ? "yes" : "no");

  if (options_done) {
    printf("Argv leftovers: ");
      for (; *argv; argv++) {
	printf ("%s ", *argv);
      }
    printf("\n");
  }
}

/*
 *----------------------------------------------------------------------
 *
 * Soar_DefineInterps --
 *
 *	The function parses an argc/argv list of options to define
 *      a set of interpreters.
 *
 * Results:
 *	A standard Tcl completion code.
 *
 * Side effects:
 *	Defines the indicated interpreters.
 *
 *----------------------------------------------------------------------
 */

int
Soar_DefineInterps (argc, argv, from_interp)
     int * argc;                      /* Number of argv items */
     char ** argv;                    /* Command line options */
     Tcl_Interp * from_interp;        /* Interp from which this function 
				       * was invoked.         */
{
  int interp_type = SOAR_ARGV_END;    /* Desired interp type to create */
  int options_done = 0;		      /* Argv completely scanned? */
  static int no_interp_created = 1;   /* Any interps created?  This is
				       * static so that it applies across
				       * repeated invocations. */
  Tcl_Interp * new_interp;            /* A newly created interp */
  int i;                              /* Loop index */
  int srcIndex = 1;                   /* Location from which to read next */
                                      /* argument from argv. */
  int dstIndex = 1;	              /* Index into argv to which next unused
				       * argument should be copied (never 
                                       * greater than srcIndex). */
  /*
   * Parse command-line arguments.
   */

  InitOptions(&options);

  (*argc)--;

  while (!options_done) 
    {
      int parse_result;

      parse_result = Soar_ParseArgv(argc, argv, argTable, 0, 
				    &srcIndex, &dstIndex,
				    &interp_type, &options_done);

      {
	char * displayString;

	displayString = getenv("DISPLAY");

	if (options.tk_enabled != 0)
	  {
	    if (    (options.display == NULL)
		&&  (   (displayString == NULL) 
		     || (!strcmp(displayString, ""))))
	      {
		options.tk_enabled = 0;
#if USE_TK
		fprintf (stderr,
			 "Warning: env. var DISPLAY not set -- disabling Tk.\n");
#endif
	      }
	  }
      }

      if (options.verbose) {
	PrintOptions(argv, &options, interp_type, options_done);
      }

      if (parse_result == TCL_ERROR)
	{
	  return TCL_ERROR;
	}

      if (   options_done 
	  && no_interp_created 
	  && (interp_type == SOAR_ARGV_END))
	{
	  /* Create a single Soar agent by default on startup */

	  if (   Soar_CreateInterpreter("soar", SOAR_ARGV_AGENT, &options,
					process_name, tty, &new_interp)
	      != TCL_OK)
	    {
	      Tcl_AppendResult(from_interp,
		       "Error creating interpreter named \"soar\".\n",
			       (char *) NULL);
	      return TCL_ERROR;
	    }

	  {
	    RegisteredInterp * riPtr;

	    riPtr = Soar_GetRegisteredInterpByInterp(new_interp);
	    if (!strcmp(riPtr->name, "soar"))
	      {
		Tcl_SetVar(new_interp, "tcl_prompt1", 
		   "puts -nonewline stdout {soar> }; log -add {soar> }",
			   TCL_GLOBAL_ONLY);
	      }
	  }

	  no_interp_created = 0;
	}
      else			/* We have a list of interpreters to create */
	{
	  if (options.names.start_argv == 0)
	    {
	      sprintf (from_interp->result, 
		       "Error: No names given in interp list.\n");
	      return TCL_ERROR;
	    }
	  else
	    {
	      for (i = options.names.start_argv; 
		   i <= options.names.end_argv; 
		   i++) 
		{
		  char * name = argv[i];

		  if (Soar_NameInRegistry(name))
		    {
		      sprintf (from_interp->result,
			       "Ignoring duplicate interpreter name: %s\n",
			       name);
		      return TCL_ERROR;
		    }

		  if (   Soar_CreateInterpreter(name, interp_type, &options,
						process_name, tty, &new_interp)
		      == TCL_ERROR)
		    {
		      Tcl_AppendResult(from_interp,
				       "\nError creating interpreter ",
				       name, ".\n", (char *) NULL);
		      return TCL_ERROR;
		    }
		  else
		    {
		      no_interp_created = 0;
		    }
		}

	      /* Erase the list of names to reset for next iteration */
	      options.names.start_argv = 0;
	      options.names.end_argv   = 0;
	    }
	}
    }
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_MainLoop --
 *
 *	This is the main loop for a Tcl-based shell that reads
 *	Tcl commands from standard input.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Can be almost arbitrary, depending on what the Tcl commands do.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_MainLoop (void)
{
  static int gotPartial = 0;
  int code;
  char buffer[1000];
  char * cmd;
  /*
   * Process commands from stdin until there's an end-of-file.
   */

  gotPartial = 0;
  Tcl_DStringInit(&command);
  while (1) {
    clearerr(stdin);
    if (tty) {
      char *promptCmd;

      promptCmd = Tcl_GetVar(interp,
			     gotPartial ? "tcl_prompt2" : "tcl_prompt1", 
			     TCL_GLOBAL_ONLY);
      if (promptCmd == NULL) {
      defaultPrompt:
	if (!gotPartial) {
	  fputs("% ", stdout);
	  if (mainInterp->agentPtr != NULL) {
	    Soar_Log(soar_agent, "% ");
	  }
	}
      } else {
	code = Tcl_Eval(interp, promptCmd);
	if (code != TCL_OK) {
	  fprintf(stderr, "%s\n", interp->result);
	  Tcl_AddErrorInfo(interp,
			   "\n    (script that generates prompt)");
	  goto defaultPrompt;
	}
      }
      fflush(stdout);
    }
    if (fgets(buffer, 1000, stdin) == NULL) {
      if (ferror(stdin)) {
	if (errno == EINTR) {
	  if (tcl_AsyncReady) {
	    (void) Tcl_AsyncInvoke((Tcl_Interp *) NULL, 0);
	  }
	  clearerr(stdin);
	} else {
	  goto done;
	}
      } else {
	if (!gotPartial) {
	  goto done;
	}
      }
      buffer[0] = 0;
    }
    cmd = Tcl_DStringAppend(&command, buffer, -1);
    if ((buffer[0] != 0) && !Tcl_CommandComplete(cmd)) {
      gotPartial = 1;
      continue;
    }

    gotPartial = 0;
    if (mainInterp->agentPtr != NULL) {
      Soar_Log(soar_agent, cmd);
    }
    code = Tcl_RecordAndEval(interp, cmd, 0);
    Tcl_DStringFree(&command);
    if (code != TCL_OK) {
      fprintf(stderr, "%s\n", interp->result);
    } else if (tty && (*interp->result != 0)) {
      printf("%s\n", interp->result);
      if (mainInterp->agentPtr != NULL) {
	Soar_Log(soar_agent, interp->result);
	Soar_Log(soar_agent, "\n");
      }
    }
  }

 done:
  return;
}
